import typing as tp
import logging

logger = logging.getLogger(__name__)


def raiseErrorDialog(title: str = 'Error',
                     message: tp.Optional[str] = None,
                     exception: tp.Optional[Exception] = None,
                     breakpointOnButton: tp.Optional[str] = 'Open',
                     buttons: tp.Optional[tp.Tuple[str, ...]] = None) -> str:

    if buttons is None:
        if __debug__ and breakpointOnButton is not None:
            buttons = ('Ignore', breakpointOnButton, 'Abort')
        else:
            buttons = ('Ignore', 'Abort')

    # lazy import to only import if error occurs (since caller may not otherwise use any GUI components)
    from PySide6 import QtWidgets, QtCore
    from .. import exceptionToStr

    if message is None:
        if exception is not None:
            message = exceptionToStr(exception)
        else:
            message = 'Error'

    if QtCore.QCoreApplication.instance() is None:
        if False:
            logger.error('Requested error dialog, but Qt is not running. Raising error immediately.')
            if exception is not None:
                raise exception
            else:
                raise RuntimeError()
        else:
            logger.info('Requested error dialog, but Qt is not running. Creating Qt app to run here.')
            try:
                import pyqtgraph as pg
            except ModuleNotFoundError as e:
                logger.error('Unable to display dialog, will raise instead.\n"%s": %s' % (title, message))
                raise exception if exception is not None else RuntimeError()

            app = pg.mkQApp()
            kwargs = dict(title=title, message=message, exception=exception, buttons=buttons)
            return raiseErrorDialog(**kwargs)


    logger.error('Displaying error dialog "%s": %s' % (title, message))

    msgBox = QtWidgets.QMessageBox()
    msgBox.setWindowTitle(title)
    msgBox.setText(message)

    btns = QtWidgets.QMessageBox.NoButton | QtWidgets.QMessageBox.NoButton
    for btnKey in buttons:
        btns |= getattr(QtWidgets.QMessageBox, btnKey)  # assume keys are enum keys

    msgBox.setStandardButtons(btns)
    msgBox.setDefaultButton(getattr(QtWidgets.QMessageBox, buttons[0]))  # use first button as the default
    msgBox.setWindowState(
        (msgBox.windowState() & ~QtCore.Qt.WindowState.WindowMinimized)
        | QtCore.Qt.WindowState.WindowActive)
    ret = msgBox.exec()

    retKey = None
    for btnKey in buttons:
        if ret == getattr(QtWidgets.QMessageBox, btnKey):
            retKey = btnKey
            break

    assert retKey is not None

    if retKey == 'Abort':
        if exception is not None:
            raise exception
        else:
            raise RuntimeError('%s: %s' % (title, message))
    elif __debug__ and breakpointOnButton is not None and retKey == breakpointOnButton:
        breakpoint()
    else:
        logger.info('User selected "%s"' % retKey)
        return retKey


async def asyncTryAndRaiseDialogOnError(fn: tp.Callable, *args, **kwargs):
    try:
        await fn(*args, **kwargs)
    except Exception as e:
        raiseErrorDialog(exception=e)



import numpy as np
import pytest
import pytransform3d.rotations as ptr
import pytransform3d.transformations as ptt

from NaviNIBS.util.numpy import array_equalish
from NaviNIBS.util.Transforms import transformToString, stringToTransform, composeTransform, invertTransform, applyTransform, estimateAligningTransform


@pytest.fixture
def transf1():
    transf = np.eye(4)
    transf[:3, :3] = ptr.active_matrix_from_extrinsic_euler_xyx([0, 1, 2])
    transf[:3, 3] = np.asarray([1, 2, 3])
    return transf


@pytest.fixture
def transf2():
    transf = np.eye(4)
    transf[:3, :3] = ptr.active_matrix_from_extrinsic_euler_xyx([3, 2, 1])
    transf[:3, 3] = np.asarray([4, 5, 6])
    return transf


@pytest.fixture(params=[0, 1])
def transf(transf1, transf2, request):
    match request.param:
        case 0:
            return transf1
        case 1:
            return transf2
        case _:
            raise NotImplementedError()


@pytest.fixture
def pts1():
    return np.asarray([10, 11, 12])


@pytest.fixture
def pts2(pts1):
    return pts1[np.newaxis, :]


@pytest.fixture
def pts3():
    return np.random.rand(20, 3)


@pytest.fixture(params=[0, 1, 2])
def pts(pts1, pts2, pts3, request):
    match request.param:
        case 0:
            return pts1
        case 1:
            return pts2
        case 2:
            return pts3
        case _:
            raise NotImplementedError()


def test_transfToFromString(transf):
    transfStr = transformToString(transf)
    assert isinstance(transfStr, str)
    transf_roundtrip = stringToTransform(transfStr)
    assert isinstance(transf_roundtrip, np.ndarray)
    assert array_equalish(transf, transf_roundtrip)


def test_invertTransform(transf):
    assert array_equalish(np.linalg.pinv(transf), invertTransform(transf))
    assert array_equalish(transf, invertTransform(invertTransform(transf)))


def test_composeTransform(transf):
    assert array_equalish(transf, composeTransform(transf[:3, :3], transf[:3, 3]))
    tmp = transf.copy()
    tmp[:3, 3] = 0
    assert array_equalish(tmp, composeTransform(transf[:3, :3]))


def test_applyTransform(transf1, transf2, pts):
    transfPts_a = applyTransform(transf2, applyTransform(transf1, pts))
    transfPts_b = applyTransform([transf1, transf2], pts)
    assert array_equalish(transfPts_a, transfPts_b)
    untransfPts = applyTransform(np.linalg.pinv(transf1), applyTransform(np.linalg.pinv(transf2), transfPts_a))
    assert array_equalish(pts, untransfPts)


@pytest.mark.parametrize("method", ['kabsch-svd'])
def test_estimateAligningTransf(transf, pts, method):
    transfPts = applyTransform(transf, pts)
    if pts.ndim < 2 or pts.shape[1] != 3 or pts.shape[0] < 3:
        with pytest.raises(ValueError):
            estTransf = estimateAligningTransform(pts, transfPts, method=method)
        return
    else:
        estTransf = estimateAligningTransform(pts, transfPts, method=method)
    assert array_equalish(estTransf, transf)
    transfPts2 = applyTransform(estTransf, pts)
    assert array_equalish(transfPts, transfPts2)


def test_estimateAligningTransf_weighted(transf, pts):
    transfPts = applyTransform(transf, pts)
    weights = np.ones((pts.shape[0],))
    weights[0] = 100
    if pts.ndim < 2 or pts.shape[1] != 3 or pts.shape[0] < 3:
        with pytest.raises(ValueError):
            estTransf = estimateAligningTransform(pts, transfPts, weights=weights)
        return
    else:
        estTransf = estimateAligningTransform(pts, transfPts, weights=weights)
    assert array_equalish(estTransf, transf)
    transfPts2 = applyTransform(estTransf, pts)
    assert array_equalish(transfPts, transfPts2)




from __future__ import annotations

import attrs
import contextlib
import datetime
import logging
import numpy as np
import pandas as pd
from pprint import pformat
import typing as tp

from NaviNIBS.util.numpy import array_equalish, attrsWithNumpyAsDict, attrsWithNumpyFromDict

from NaviNIBS.Navigator.Model.GenericCollection import GenericCollection, GenericCollectionDictItem


logger = logging.getLogger(__name__)


Timestamp = pd.Timestamp


@attrs.define
class Sample(GenericCollectionDictItem[str]):
    """
    Represents a single recorded sample
    """
    _timestamp: Timestamp
    _coilToMRITransf: tp.Optional[np.ndarray] = None
    _targetKey: tp.Optional[str] = None
    """
    Key of target that was active at the time sample was collected. Note that if target info (e.g. coordinates) changed later,
    this should automatically be updated to refer to a "historical" target with info matching that of the target at the time
    the sample was created.
    """
    _coilKey: tp.Optional[str] = None
    """
    Key of coil tool that was active at the time sample was collected.
    """

    _isVisible: bool = True
    _isSelected: bool = False
    _color: tp.Optional[str] = None

    _metadata: dict[str, tp.Any] = attrs.field(factory=dict)
    """
    For storing misc metadata, such as information about the trigger event that initiated this sample.
    
    Values should be JSON-serializable.
    """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, newTimestamp: Timestamp):
        if self._timestamp == newTimestamp:
            return
        self.sigItemAboutToChange.emit(self.key, ['timestamp'])
        self._timestamp = newTimestamp
        self.sigItemChanged.emit(self.key, ['timestamp'])

    @property
    def coilToMRITransf(self):
        return self._coilToMRITransf

    @coilToMRITransf.setter
    def coilToMRITransf(self, newTransf: tp.Optional[np.ndarray]):
        if array_equalish(self._coilToMRITransf, newTransf):
            return
        self.sigItemAboutToChange.emit(self.key, ['coilToMRITransf'])
        self._coilToMRITransf = newTransf
        self.sigItemChanged.emit(self.key, ['coilToMRITransf'])

    @property
    def hasTransf(self):
        return self._coilToMRITransf is not None

    @property
    def targetKey(self):
        return self._targetKey

    @targetKey.setter
    def targetKey(self, newKey: tp.Optional[str]):
        if self._targetKey == newKey:
            return
        self.sigItemAboutToChange.emit(self.key, ['targetKey'])
        self._targetKey = newKey
        self.sigItemChanged.emit(self.key, ['targetKey'])

    @property
    def coilKey(self):
        return self._coilKey

    @coilKey.setter
    def coilKey(self, newKey: tp.Optional[str]):
        if self._coilKey == newKey:
            return
        self.sigItemAboutToChange.emit(self.key, ['coilKey'])
        self._coilKey = newKey
        self.sigItemChanged.emit(self.key, ['coilKey'])

    @property
    def isVisible(self):
        return self._isVisible

    @isVisible.setter
    def isVisible(self, isVisible: bool):
        if self._isVisible == isVisible:
            return
        self.sigItemAboutToChange.emit(self.key, ['isVisible'])
        self._isVisible = isVisible
        self.sigItemChanged.emit(self.key, ['isVisible'])

    @property
    def isSelected(self):
        return self._isSelected

    @isSelected.setter
    def isSelected(self, isSelected: bool):
        if self._isSelected == isSelected:
            return
        self.sigItemAboutToChange.emit(self.key, ['isSelected'])
        self._isSelected = isSelected
        self.sigItemChanged.emit(self.key, ['isSelected'])

    @property
    def color(self):
        return self._color

    @property
    def metadata(self):
        """
        Note: if needing to modify the result, make sure to do within the `changingMetadata` context manager, like:
                with sample.changingMetadata() as metadata:
                    metadata['foo'] = 'bar'
        """
        return self._metadata

    @contextlib.contextmanager
    def changingMetadata(self):
        self.sigItemAboutToChange.emit(self.key, ['metadata'])
        yield self._metadata
        self.sigItemChanged.emit(self.key, ['metadata'])

    def __str__(self):
        return pformat(self.asDict())

    def asDict(self) -> tp.Dict[str, tp.Any]:
        d = attrsWithNumpyAsDict(self, npFields=('coilToMRITransf',))

        d['timestamp'] = d['timestamp'].isoformat(timespec='microseconds')  # default may include nanoseconds, which can break when calling with fromisoformat on later import

        return d

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any]):

        d['timestamp'] = Timestamp.fromisoformat(d['timestamp'])

        return attrsWithNumpyFromDict(cls, d, npFields=('coilToMRITransf',))


@attrs.define
class Samples(GenericCollection[str, Sample]):
    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def getUniqueSampleKey(self,
                           baseStr: str = 'Sample ',
                           startAtIndex: tp.Optional[int] = None,
                           timestamp: tp.Optional[Timestamp] = None) -> str:
        """
        Get a key not already used by any current samples, presumably to use for a new sample.

        If a timestamp is specified, will format the key using the timestamp. Otherwise will be based on
        an index and length of samples.
        """

        key = baseStr
        if timestamp is None:
            if startAtIndex is None:
                index = len(self._items)
            else:
                index = startAtIndex
            while True:
                key = f'{baseStr}{index}'
                if key not in self.keys():
                    return key
                index += 1
        else:
            removeNChars = [7, 3, 0]
            iAttempt = 0
            while iAttempt < len(removeNChars):
                key = baseStr + timestamp.strftime('%y.%m.%d %H:%M:%S.%f')
                if removeNChars[iAttempt] > 0:
                    key = key[:-removeNChars[iAttempt]]
                if key.endswith('.'):
                    key = key[:-1]
                if key not in self.keys():
                    return key
                iAttempt += 1

            if startAtIndex is None:
                startAtIndex = 2
            return self.getUniqueSampleKey(baseStr=key, startAtIndex=startAtIndex)

    def setWhichSamplesVisible(self, visibleKeys: list[str]):
        self.setAttribForItems(self.keys(), dict(isVisible=[key in visibleKeys for key in self.keys()]))

    def setWhichSamplesSelected(self, selectedKeys: tp.List[str]):
        self.setAttribForItems(self.keys(), dict(isSelected=[key in selectedKeys for key in self.keys()]))

    @classmethod
    def fromList(cls, itemList: list[dict[str, tp.Any]]) -> Samples:
        items = {}
        for itemDict in itemList:
            items[itemDict['key']] = Sample.fromDict(itemDict)

        return cls(items=items)


def getSampleTimestampNow() -> Timestamp:
    return pd.Timestamp.now()



from __future__ import annotations

import shutil

import attrs
from copy import deepcopy
from datetime import datetime
import nibabel as nib
import jsbeautifier
import json
import logging
import numpy as np
import os
import pandas as pd
import pyvista as pv
import tempfile
import typing as tp
from typing import ClassVar

import NaviNIBS
from NaviNIBS.Navigator.Model.MiscSettings import MiscSettings
from NaviNIBS.Navigator.Model.MRI import MRI
from NaviNIBS.Navigator.Model.HeadModel import HeadModel
from NaviNIBS.Navigator.Model.CoordinateSystems import CoordinateSystems, CoordinateSystem
from NaviNIBS.Navigator.Model.Targets import Targets, Target
from NaviNIBS.Navigator.Model.Samples import Samples, Sample
from NaviNIBS.Navigator.Model.SubjectRegistration import SubjectRegistration, Fiducial
from NaviNIBS.Navigator.Model.Tools import Tools, Tool, CoilTool, Pointer, SubjectTracker, CalibrationPlate
from NaviNIBS.Navigator.Model.Triggering import TriggerSources, TriggerSource
from NaviNIBS.Navigator.Model.DigitizedLocations import DigitizedLocations, DigitizedLocation
from NaviNIBS.Navigator.Model.DockWidgetLayouts import DockWidgetLayouts
from NaviNIBS.Navigator.Model.Addons import Addons, Addon
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.numpy import array_equalish


logger = logging.getLogger(__name__)


@attrs.define
class MNIRegistration:
    sigTransformChanged: Signal = attrs.field(init=False, factory=Signal)
    # TODO


@attrs.define
class Session:
    """
    Primary data model for a NaviNIBS session. Contains all session-specific data and methods for saving/loading.
    """
    _filepath: str  # path to compressed session file
    _subjectID: tp.Optional[str] = attrs.field(default=None)
    _sessionID: tp.Optional[str] = None
    _miscSettings: MiscSettings = attrs.field(factory=MiscSettings)
    _MRI: MRI = attrs.field(factory=MRI)
    _headModel: HeadModel = attrs.field(factory=HeadModel)
    _coordinateSystems: CoordinateSystems = attrs.field(factory=CoordinateSystems)
    _subjectRegistration: SubjectRegistration = attrs.field(factory=SubjectRegistration)
    _targets: Targets = attrs.field(factory=Targets)
    _tools: Tools = attrs.field(default=None)
    _samples: Samples = attrs.field(factory=Samples)
    _digitizedLocations: DigitizedLocations = attrs.field(factory=DigitizedLocations)
    _triggerSources: TriggerSources = attrs.field(factory=TriggerSources)
    _dockWidgetLayouts: DockWidgetLayouts = attrs.field(factory=DockWidgetLayouts)
    _addons: Addons = attrs.field(factory=Addons)

    _dirtyKeys: tp.Set[str] = attrs.field(init=False, factory=set)  # session info parts that have changed since last save
    _dirtyKeys_autosave: tp.Set[str] = attrs.field(init=False, factory=set)  # session info parts that have changed since last autosave
    _compressedFileIsDirty: bool = True

    _sessionConfigFilename: ClassVar[str] = 'SessionConfig'
    _lastAutosaveFilenamePrefix: tp.Optional[str] = None
    _latestConfigFormatVersion: ClassVar[str] = '0.0.2'
    """
    Config format version history:
    - 0.0.1: Initial format, went through a number of iterations
    - 0.0.2:
        - Switched from dict to list of trigger sources
        - Switched fiducial format
    """
    _unpackedSessionDir: tp.Optional[tp.Union[tempfile.TemporaryDirectory, str]] = attrs.field(default=None)

    _beautifier: jsbeautifier.Beautifier | None = attrs.field(init=False, default=None)

    sigInfoChanged: Signal = attrs.field(init=False, factory=lambda: Signal((tp.Optional[list[str]])))
    """
    Includes list of keys in info that were changed. If list is None, subscribers should assume that all info changed.
    """
    sigDirtyKeysChanged: Signal[()] = attrs.field(init=False, factory=Signal)

    def __attrs_post_init__(self):
        if self._unpackedSessionDir is None:
            self._unpackedSessionDir = self.getTempUnpackDir()

        if not os.path.isdir(self.unpackedSessionDir):
            logger.debug('Creating dir for unpacking session at {}'.format(self.unpackedSessionDir))
            os.makedirs(self.unpackedSessionDir)

        if self._tools is None:
            self._tools = Tools(sessionPath=self._filepath)

        self.sigInfoChanged.connect(lambda *args: self.flagKeyAsDirty('info'))
        self.miscSettings.sigAttribsChanged.connect(lambda *args: self.flagKeyAsDirty('miscSettings'))
        self.MRI.sigFilepathChanged.connect(lambda: self.flagKeyAsDirty('MRI'))
        self.MRI.sigManualClimChanged.connect(lambda *args: self.flagKeyAsDirty('MRI'))
        self.headModel.sigFilepathChanged.connect(lambda: self.flagKeyAsDirty('headModel'))
        self.subjectRegistration.fiducials.sigItemsChanged.connect(lambda *args: self.flagKeyAsDirty('subjectRegistration'))
        self.subjectRegistration.sampledHeadPoints.sigHeadpointsChanged.connect(lambda *args: self.flagKeyAsDirty('subjectRegistration'))
        self.subjectRegistration.sampledHeadPoints.sigAttribsChanged.connect(lambda *args: self.flagKeyAsDirty('subjectRegistration'))
        self.subjectRegistration.sigTrackerToMRITransfChanged.connect(lambda: self.flagKeyAsDirty('subjectRegistration'))
        self.targets.sigItemsChanged.connect(lambda targetKeys, attribKeys: self.flagKeyAsDirty('targets'))
        self.samples.sigItemsChanged.connect(lambda sampleTimestamps, attribKeys: self.flagKeyAsDirty('samples'))
        self.tools.sigItemsChanged.connect(lambda *args: self.flagKeyAsDirty('tools'))
        self.tools.sigPositionsServerInfoChanged.connect(lambda *args: self.flagKeyAsDirty('tools'))
        self.triggerSources.sigItemsChanged.connect(lambda *args: self.flagKeyAsDirty('triggerSources'))
        self._dockWidgetLayouts.sigItemsChanged.connect(lambda *args: self.flagKeyAsDirty('dockWidgetLayouts'))
        self.addons.sigItemsAboutToChange.connect(self._onAddonsAboutToChange)
        self.addons.sigItemsChanged.connect(self._onAddonsChanged)
        self.targets.sigItemKeyChanged.connect(self._updateSamplesForNewTargetKey)
        self.targets.sigItemsAboutToChange.connect(self._onTargetsAboutToChange, priority=1)  # use higher priority to make sure we handle adding historical samples before notifying GUIs of this change
        self.targets.sigItemsChanged.connect(self._onTargetsChanged)
        self.samples.sigItemsChanged.connect(self._onSamplesChanged)
        self.coordinateSystems.sigItemsChanged.connect(self._onCoordinateSystemsChanged)
        self.digitizedLocations.sigItemsChanged.connect(lambda *args: self.flagKeyAsDirty('digitizedLocations'))

        self.coordinateSystems.session = self
        self.targets.session = self

        # TODO

    @property
    def subjectID(self):
        return self._subjectID

    @subjectID.setter
    def subjectID(self, newVal: str):
        if newVal != self._subjectID:
            self._subjectID = newVal
            self.sigInfoChanged.emit(['subjectID'])

    @property
    def sessionID(self):
        return self._sessionID

    @sessionID.setter
    def sessionID(self, newVal: str):
        if newVal != self._sessionID:
            self._sessionID = newVal
            self.sigInfoChanged.emit(['sessionID'])

    @property
    def filepath(self):
        return self._filepath

    @filepath.setter
    def filepath(self, newVal: str):
        if newVal != self._filepath:
            self._filepath = newVal
            self.sigInfoChanged.emit(['filepath'])
            self.tools.sessionPath = self._filepath

    @property
    def miscSettings(self):
        return self._miscSettings

    @property
    def MRI(self):
        return self._MRI

    @property
    def headModel(self):
        return self._headModel

    @property
    def coordinateSystems(self):
        return self._coordinateSystems

    @property
    def digitizedLocations(self):
        return self._digitizedLocations

    @property
    def subjectRegistration(self):
        return self._subjectRegistration

    @property
    def targets(self):
        return self._targets

    @property
    def samples(self):
        return self._samples

    @property
    def tools(self):
        return self._tools

    @property
    def triggerSources(self):
        return self._triggerSources

    @property
    def dockWidgetLayouts(self):
        return self._dockWidgetLayouts

    @property
    def addons(self):
        return self._addons

    @property
    def compressedFileIsDirty(self):
        return len(self._dirtyKeys) > 0 or self._compressedFileIsDirty

    @property
    def unpackedSessionDir(self):
        if isinstance(self._unpackedSessionDir, tempfile.TemporaryDirectory):
            return self._unpackedSessionDir.name
        else:
            return self._unpackedSessionDir

    @unpackedSessionDir.setter
    def unpackedSessionDir(self, newUnpackedSessionDir: str):
        if newUnpackedSessionDir == self.unpackedSessionDir:
            # no change
            return

        # rather than just changing directory and resaving, copy contents of entire previous directory
        #  (to make sure we bring any dependencies like session-specific images with us)
        logger.info('Copying contents of previous session dir to new location')
        shutil.copytree(self.unpackedSessionDir, newUnpackedSessionDir, dirs_exist_ok=True)
        # note that any existing files in new folder that don't also exist in old folder will remain in place (not be deleted)
        logger.debug('Done copying')

        self._unpackedSessionDir = newUnpackedSessionDir

    @property
    def beautifier(self):
        """
        Cached instance of json beautifier. This saves some time compared to initializing with each dumps call.
        """

        if self._beautifier is None:
            opts = jsbeautifier.default_options()
            opts.indent_size = 2
            self._beautifier = jsbeautifier.Beautifier(opts)

        return self._beautifier

    def flagKeyAsDirty(self, key: str):
        self._dirtyKeys.add(key)
        self._dirtyKeys_autosave.add(key)
        self.sigDirtyKeysChanged.emit()

    @property
    def dirtyKeys(self):
        """
        Result should not be modified.
        """
        return self._dirtyKeys

    def saveToUnpackedDir(self, saveDirtyOnly: bool = True, asAutosave: bool = False):

        if asAutosave:
            keysToSave = self._dirtyKeys_autosave.copy()
            self._dirtyKeys_autosave.clear()

            autosaveFilenamePrefix = 'autosaved-' + datetime.today().strftime('%y%m%d%H%M%S.%f') + '_'
            if True:
                # put autosaved files in an "autosaved" subdirectory to avoid cluttering session root
                autosaveFilenamePrefix = os.path.join('autosaved', autosaveFilenamePrefix)
                if not os.path.isdir(os.path.join(self.unpackedSessionDir, 'autosaved')):
                    os.mkdir(os.path.join(self.unpackedSessionDir, 'autosaved'))
        else:
            keysToSave = self._dirtyKeys.copy()
            self._dirtyKeys.clear()
            self._dirtyKeys_autosave.clear()
            self.sigDirtyKeysChanged.emit()
            self._lastAutosaveFilenamePrefix = None
            autosaveFilenamePrefix = ''

        if saveDirtyOnly and len(keysToSave) == 0:
            logger.debug('Nothing to save')
            return

        self._compressedFileIsDirty = True

        if asAutosave:
            configPath = os.path.join(self.unpackedSessionDir, autosaveFilenamePrefix + self._sessionConfigFilename + '.json')
            if self._lastAutosaveFilenamePrefix is not None:
                prevConfigPath = os.path.join(self.unpackedSessionDir, self._lastAutosaveFilenamePrefix + self._sessionConfigFilename + '.json')
            else:
                prevConfigPath = os.path.join(self.unpackedSessionDir, self._sessionConfigFilename + '.json')
            self._lastAutosaveFilenamePrefix = autosaveFilenamePrefix
        else:
            configPath = os.path.join(self.unpackedSessionDir, self._sessionConfigFilename + '.json')
            prevConfigPath = configPath

        if os.path.exists(prevConfigPath):
            with open(prevConfigPath, 'r') as f:
                config = json.load(f)
            assert config['formatVersion'] == self._latestConfigFormatVersion
        else:
            config = dict(formatVersion=self._latestConfigFormatVersion)

        config['softwareVersion'] = NaviNIBS.__version__

        def saveConfigPartToFileIfNeeded(key: str, getWhatToDump: tp.Callable[[], tp.Any]):
            if key in keysToSave or not saveDirtyOnly:
                logger.debug(f'Writing {key} info')
                upperKey = key[0].upper() + key[1:]
                configFilename_part = autosaveFilenamePrefix + self._sessionConfigFilename + '_' + upperKey + '.json'
                outputPath = os.path.join(self.unpackedSessionDir, configFilename_part)
                toDump = getWhatToDump()
                if len(toDump) == 0:
                    # delete output if it already exists
                    if os.path.exists(outputPath):
                        os.remove(outputPath)
                    if key in config:
                        del config[key]
                else:
                    config[key] = configFilename_part
                    toWrite = self._prettyJSONDumps(toDump)
                    with open(outputPath, 'w') as f:
                        f.write(toWrite)
                keysToSave.discard(key)

        if 'info' in keysToSave or not saveDirtyOnly:
            logger.debug('Writing session info')
            infoFields = ('filepath', 'subjectID', 'sessionID')
            for field in infoFields:
                config[field] = getattr(self, field)
            keysToSave.discard('info')

        otherPathsRelTo = self.filepath

        saveConfigPartToFileIfNeeded('miscSettings', lambda: self.miscSettings.asDict())

        saveConfigPartToFileIfNeeded('MRI', lambda: self.MRI.asDict(filepathRelTo=otherPathsRelTo))

        saveConfigPartToFileIfNeeded('headModel', lambda: self.headModel.asDict(filepathRelTo=otherPathsRelTo))

        saveConfigPartToFileIfNeeded('coordinateSystems', lambda: self.coordinateSystems.asList())

        saveConfigPartToFileIfNeeded('digitizedLocations', lambda: self.digitizedLocations.asList())

        saveConfigPartToFileIfNeeded('subjectRegistration', lambda: self.subjectRegistration.asDict())
        # TODO: maybe save contents of potentially larger *History fields to separate file(s)

        saveConfigPartToFileIfNeeded('targets', lambda: self.targets.asList())

        saveConfigPartToFileIfNeeded('samples', lambda: self.samples.asList())

        saveConfigPartToFileIfNeeded('tools', lambda: self.tools.asList())

        thisKey = 'triggerSources'
        if thisKey in keysToSave or not saveDirtyOnly:
            logger.debug(f'Writing {thisKey} info')
            config[thisKey] = self.triggerSources.asList()
            if len(config[thisKey]) == 0:
                 del config[thisKey]
            keysToSave.discard(thisKey)

        saveConfigPartToFileIfNeeded('dockWidgetLayouts', lambda: self.dockWidgetLayouts.asList())

        # TODO: loop through any addons to give them a chance to save to config as needed

        thisKey = 'addons'
        if thisKey in keysToSave or not saveDirtyOnly:
            logger.debug(f'Writing {thisKey} info')
            config[thisKey] = self.addons.asList(
                unpackedSessionDir=self.unpackedSessionDir,
                filenamePrefix=autosaveFilenamePrefix + self._sessionConfigFilename + '_')
            if len(config[thisKey]) == 0:
                 del config[thisKey]
            keysToDiscard = {thisKey}
            for key in keysToSave:
                if key.startswith('addon.'):
                    keysToDiscard.add(key)
            keysToSave -= keysToDiscard
        else:
            # write config for changed addons only, not all addons
            for addonKey, addon in self.addons.items():
                flagKey = f'addon.{addonKey}'
                if flagKey in keysToSave:
                    configFilename_addon = addon.writeConfig(
                        unpackedSessionDir=self.unpackedSessionDir,
                        filenamePrefix=autosaveFilenamePrefix + self._sessionConfigFilename + '_')

                    altConfigFilename_addon = addon.getConfigFilename(
                        unpackedSessionDir=self.unpackedSessionDir,
                        filenamePrefix=self._sessionConfigFilename + '_',  # no autosave prefix
                    )
                    if altConfigFilename_addon != configFilename_addon:
                        if not altConfigFilename_addon in config['addons']:
                            for prevAddonFilename in config['addons']:
                                if prevAddonFilename.endswith(altConfigFilename_addon):
                                    altConfigFilename_addon = prevAddonFilename
                                    break
                        if altConfigFilename_addon in config['addons']:
                            # update config to use new filename
                            config['addons'][config['addons'].index(altConfigFilename_addon)] = configFilename_addon
                    assert configFilename_addon in config['addons']

                    keysToSave.discard(flagKey)

        # TODO: save other fields
        assert len(keysToSave) == 0

        with open(configPath, 'w') as f:
            if False:
                json.dump(config, f)
            else:
                f.write(self._prettyJSONDumps(config))
            logger.debug('Wrote updated session config')

    def _updateSamplesForNewTargetKey(self, fromKey: str, toKey: str):
        # update any referenced target IDs in samples to use the new key
        target = self.targets[toKey]
        if not target.mayBeADependency:
            # assume this flag would be set if any samples reference the target,
            # so no need to check samples
            return
        for sampleKey, sample in self.samples.items():
            if sample.targetKey == fromKey:
                logger.debug(f'Updating associated targetKey for {sampleKey} from {fromKey} to {toKey}')
                sample.targetKey = toKey

    def _onTargetsAboutToChange(self, keys: list[str], changingAttrs: tp.Optional[list[str]] = None):
        for targetKey in keys:
            if targetKey not in self._targets:
                # probably a completely new target
                continue

            target = self._targets[targetKey]
            if not target.mayBeADependency:
                continue
            if changingAttrs is None or any(x in changingAttrs for x in ('targetCoord',
                                                                         'entryCoord',
                                                                         'angle',
                                                                         'depthOffset',
                                                                         'coilToMRITransf')):
                # targeting info is about to change, and this target may be referenced by samples
                # so make a copy to keep in history
                logger.info('Creating historical copy of target')
                historicalTarget = Target.fromDict(deepcopy(target.asDict()), session=self)
                historicalTarget.key = targetKey + ' pre ' + datetime.today().strftime('%y%m%d%H%M%S.%f')
                historicalTarget.isHistorical = True
                target.mayBeADependency = False
                self._targets.addItem(historicalTarget)
                self._updateSamplesForNewTargetKey(fromKey=target.key, toKey=historicalTarget.key)

    def _onTargetsChanged(self, keys: list[str], changedAttrs: tp.Optional[list[str]] = None):
        pass

    def _onSamplesChanged(self, keys: list[str], changedAttrs: tp.Optional[list[str]] = None):
        if changedAttrs is None or 'targetKey' in changedAttrs:
            for sampleKey in keys:
                sample = self.samples[sampleKey]
                targetKey = sample.targetKey
                if targetKey is not None and targetKey in self._targets:
                    # mark that if there are future changes in the target, a copy may need to be kept (or this sample may need to be notified)
                    self._targets[targetKey].mayBeADependency = True

    def _onCoordinateSystemsChanged(self, keys: list[str], changedAttrs: tp.Optional[list[str]] = None):
        # only mark coordinate systems as dirty if a non-autogenerated coordSys changed
        # (changes in autogenerated coordinate systems don't require config file updates)
        isDirty = False
        for key in keys:
            if not self.coordinateSystems[key].isAutogenerated:
                isDirty = True
                break
        if isDirty:
            self.flagKeyAsDirty('coordinateSystems')

    def _onAddonsAboutToChange(self, addonKeys: list[str], changingAttrs: tp.Optional[list[str]] = None):
        for addonKey in addonKeys:
            if addonKey not in self.addons:
                # completely new addon
                self.flagKeyAsDirty('addons')

    def _onAddonsChanged(self, addonKeys: list[str], attribKeys: tp.Optional[list[str]] = None):
        for addonKey in addonKeys:
            self.flagKeyAsDirty(f'addon.{addonKey}')

    def _prettyJSONDumps(self, obj):
        return self.beautifier.beautify(json.dumps(obj))

    def saveToFile(self, updateDirtyOnly: bool = True):
        self.saveToUnpackedDir(saveDirtyOnly=updateDirtyOnly)
        if self._filepath == self._unpackedSessionDir:
            # original session file was already an unpacked dir, don't need to compress now
            logger.info('Saving to unpacked session dir only, skipping save of compressed session file.')
            return

        if not self._compressedFileIsDirty:
            logger.warning('Nothing to save')
            return

        logger.debug('Making archive')
        shutil.make_archive(
            base_name=self._filepath,
            format='zip',
            root_dir=self.unpackedSessionDir,
            base_dir='.',
        )
        shutil.move(self._filepath + '.zip', self._filepath)
        logger.debug('Done saving')

        self._compressedFileIsDirty = False

    def mergeFromFile(self, filepath: str, sections: tp.Optional[tp.List[str]] = None):
        """
        Import session elements from another file. Specify `sections` to only read a subset of elements from the file to merge, e.g. `sections=['targets']` to ignore everything but the targets section in the loaded file.
        """

        logger.info('Merge {} from file: {}'.format('all' if sections is None else sections, filepath))

        _, ext = os.path.splitext(filepath)

        if sections is not None:
            sections = sections.copy()

        if ext == '.json':
            with open(filepath, 'r') as f:
                d = json.load(f)
            # TODO: validate against schema

            if sections is None or 'targets' in sections:
                if sections is not None:
                    sections.remove('targets')
                if isinstance(d, dict):
                    assert 'targets' in d, 'Targets to import/merge should be in json with "targets" as a field in a root-level dict'
                    newTargets = Targets.fromList(d['targets'], session=self)
                else:
                    assert isinstance(d, list)
                    newTargets = Targets.fromList(d, session=self)
                self.targets.merge(newTargets)

            if sections is None or 'tools' in sections:
                if sections is not None:
                    sections.remove('tools')
                if isinstance(d, dict):
                    assert 'tools' in d, 'Tools to import/merge should be in json with "tools" as a field in a root-level dict'
                    newTools = Tools.fromList(d['tools'], sessionPath=self._filepath)
                else:
                    assert isinstance(d, list)
                    newTools = Tools.fromList(d, sessionPath=self._filepath)
                self.tools.merge(newTools)

            if sections is not None and len(sections) > 0:
                raise NotImplementedError('Merging of sections {} not implemented yet'.format(sections))
        else:
            raise NotImplementedError()  # TODO: implement more general merging of .navinibs files

    @classmethod
    def createNew(cls, filepath: str, unpackedSessionDir: tp.Optional[str] = None):
        self = cls(filepath=filepath, unpackedSessionDir=unpackedSessionDir)
        self.saveToUnpackedDir(saveDirtyOnly=False)
        return self

    @classmethod
    def loadFromFile(cls, filepath: str, unpackedSessionDir: tp.Optional[str] = None):
        if unpackedSessionDir is None:
            unpackedSessionDir = cls.getTempUnpackDir()
        logger.debug('Unpacking archive from {}\nto {}'.format(filepath, unpackedSessionDir))
        assert os.path.exists(filepath)
        shutil.unpack_archive(filepath, unpackedSessionDir, 'zip')
        logger.debug('Done unpacking')
        return cls.loadFromUnpackedDir(unpackedSessionDir=unpackedSessionDir, filepath=filepath,
                                       compressedFileIsDirty=False)

    @classmethod
    def loadFromFolder(cls, folderpath: str, **kwargs):
        return cls.loadFromUnpackedDir(unpackedSessionDir=folderpath, filepath=folderpath, **kwargs)

    @classmethod
    def loadFromUnpackedDir(cls, unpackedSessionDir: str, filepath: tp.Optional[str] = None, **kwargs):
        configPath = os.path.join(unpackedSessionDir, cls._sessionConfigFilename + '.json')
        with open(configPath, 'r') as f:
            config = json.load(f)
        # TODO: validate against schema
        assert config['formatVersion'] == cls._latestConfigFormatVersion

        kwargs['unpackedSessionDir'] = unpackedSessionDir
        kwargs['filepath'] = filepath if filepath is not None else config['filepath']
        for key in ('subjectID', 'sessionID'):
            kwargs[key] = config[key]

        otherPathsRelTo = kwargs['filepath']

        if 'miscSettings' in config:
            configFilename_miscSettings = config['miscSettings']
            with open(os.path.join(unpackedSessionDir, configFilename_miscSettings), 'r') as f:
                kwargs['miscSettings'] = MiscSettings.fromDict(json.load(f))

        if 'MRI' in config:
            configFilename_MRI = config['MRI']
            with open(os.path.join(unpackedSessionDir, configFilename_MRI), 'r') as f:
                kwargs['MRI'] = MRI.fromDict(json.load(f), filepathRelTo=otherPathsRelTo)

        if 'headModel' in config:
            configFilename_headModel = config['headModel']
            with open(os.path.join(unpackedSessionDir, configFilename_headModel), 'r') as f:
                kwargs['headModel'] = HeadModel.fromDict(json.load(f), filepathRelTo=otherPathsRelTo)

        if 'coordinateSystems' in config:
            configFilename_coordinateSystems = config['coordinateSystems']
            with open(os.path.join(unpackedSessionDir, configFilename_coordinateSystems), 'r') as f:
                kwargs['coordinateSystems'] = CoordinateSystems.fromList(json.load(f))

        if 'digitizedLocations' in config:
            configFilename_digitizedLocations = config['digitizedLocations']
            with open(os.path.join(unpackedSessionDir, configFilename_digitizedLocations), 'r') as f:
                kwargs['digitizedLocations'] = DigitizedLocations.fromList(json.load(f))

        if 'subjectRegistration' in config:
            configFilename_subjectReg = config['subjectRegistration']
            with open(os.path.join(unpackedSessionDir, configFilename_subjectReg), 'r') as f:
                kwargs['subjectRegistration'] = SubjectRegistration.fromDict(json.load(f))

        if 'targets' in config:
            configFilename_targets = config['targets']
            with open(os.path.join(unpackedSessionDir, configFilename_targets), 'r') as f:
                kwargs['targets'] = Targets.fromList(json.load(f))

        if 'samples' in config:
            configFilename_samples = config['samples']
            with open(os.path.join(unpackedSessionDir, configFilename_samples), 'r') as f:
                kwargs['samples'] = Samples.fromList(json.load(f))

        if 'tools' in config:
            configFilename_tools = config['tools']
            with open(os.path.join(unpackedSessionDir, configFilename_tools), 'r') as f:
                kwargs['tools'] = Tools.fromList(json.load(f), sessionPath=otherPathsRelTo)

        if 'triggerSources' in config:
            kwargs['triggerSources'] = TriggerSources.fromList(config['triggerSources'])

        if 'dockWidgetLayouts' in config:
            configFilename_docWidgetLayouts = config['dockWidgetLayouts']
            with open(os.path.join(unpackedSessionDir, configFilename_docWidgetLayouts), 'r') as f:
                kwargs['dockWidgetLayouts'] = DockWidgetLayouts.fromList(json.load(f))

        if 'addons' in config:
            kwargs['addons'] = Addons.fromList(config['addons'], unpackedSessionDir=unpackedSessionDir)

        # TODO: load other available fields

        logger.debug('Loaded from unpacked dir:\n{}'.format(kwargs))

        return cls(**kwargs)

    @classmethod
    def getTempUnpackDir(cls):
        return tempfile.TemporaryDirectory(prefix='NaviNIBSSession_').name



from __future__ import annotations

import asyncio

import attrs
import logging
import multiprocessing as mp
import numpy as np
import os
import pathlib
import time

import pandas as pd
import shutil
import typing as tp
from typing import ClassVar

from NaviNIBS.Devices.ToolPositionsClient import ToolPositionsClient
from NaviNIBS.Navigator.Model.Session import Session, Tool, CoilTool, SubjectTracker, Target, Sample
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.CoilOrientations import PoseMetricCalculator
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.Transforms import invertTransform, concatenateTransforms, applyTransform
from NaviNIBS.util.GUI.QFileSelectWidget import QFileSelectWidget


logger = logging.getLogger(__name__)


Transform = np.ndarray


@attrs.define(frozen=True)
class ProjectionSpecification:
    """
    Specifiers to describe projection of an orientation down the depth axis to a target plane (or sphere)
    """
    _toOrientation: str  # 'target' or 'coil'
    _toDepth: str  # if toOrientation=='coil', can be one of ['coil', 'skin', 'gm']; if toOrientation=='target', can be ['coil', 'skin', 'gm', 'target']
    _toShape: str  # 'sphere' or 'plane'


_targetingCoordinatorSingleton = None


@attrs.define
class TargetingCoordinator:
    _session: Session = attrs.field(repr=False)
    _currentTargetKey: tp.Optional[str] = None
    _currentSampleKey: tp.Optional[str] = None
    _positionsClient: ToolPositionsClient = attrs.field(factory=ToolPositionsClient)
    __cachedActiveCoilKey: str | None = attrs.field(init=False, default=None)
    __cachedActiveCoil: CoilTool | None = attrs.field(init=False, default=None)

    _currentCoilToMRITransform: tp.Optional[Transform] = attrs.field(init=False, default=None)  # relative to head tracker
    _currentPoseMetrics: PoseMetricCalculator = attrs.field(init=False, repr=False)
    _currentSamplePoseMetrics: PoseMetricCalculator = attrs.field(init=False, repr=False)

    _isOnTargetWhenDistErrorUnder: float = 1  # in mm
    _isOnTargetWhenZAngleErrorUnder: float = 2.  # in deg
    _isOnTargetWhenHorizAngleErrorUnder: float = 4.  # in deg
    _isOnTargetWhenZDistErrorBetween: tuple[float, float] = (-8, 4.)  # in mm
    """
    in mm. Uses span to allow greater error in negative region (i.e. coil closer to head than planned).
    """
    _isOnTargetMinTime: float = 0.5  # in sec, don't report being on target until after staying on target for at least this long
    _isOffTargetWhenDistErrorExceeds: float = 2.  # in mm
    _isOffTargetWhenZAngleErrorExceeds: float = 4.  # in deg
    _isOffTargetWhenHorizAngleErrorExceeds: float = 8.  # in deg
    _isOffTargetWhenZDistErrorExceeds: float = 8.  # in mm
    _isOffTargetMinTime: float = 0.1  # in sec, don't report being off target until after staying off target for at least this long
    _doMonitorOnTarget: bool = False
    _monitorOnTargetRate: float = 5.  # in Hz
    _isOnTarget: bool = attrs.field(init=False, default=False)
    _onTargetMaybeChangedAtTime: float | None = None
    _needToCheckIfOnTarget: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)
    _monitorOnTargetTask: asyncio.Task | None = attrs.field(init=False, default=None, repr=False)
    sigIsOnTargetChanged: Signal = attrs.field(init=False, factory=Signal, repr=False)  # only emitted when doMonitorOnTarget is True and isOnTarget changes

    sigActiveCoilKeyChanged: Signal = attrs.field(init=False, factory=Signal, repr=False)
    sigCurrentTargetChanged: Signal = attrs.field(init=False, factory=Signal, repr=False)
    """
    Emitted when a different target becomes current AND when an attribute of the current target changes. 
    """
    sigCurrentSampleChanged: Signal = attrs.field(init=False, factory=Signal, repr=False)
    sigCurrentCoilPositionChanged: Signal = attrs.field(init=False, factory=Signal, repr=False)
    sigCurrentSubjectPositionChanged: Signal = attrs.field(init=False, factory=Signal, repr=False)

    def __attrs_post_init__(self):
        self._positionsClient.sigLatestPositionsChanged.connect(self._onLatestPositionsChanged)
        self._session.tools.sigItemsAboutToChange.connect(self._onToolsAboutToChange)
        self._session.tools.sigItemsChanged.connect(self._onToolsChanged)
        self._session.targets.sigItemsChanged.connect(self._onSessionTargetsChanged)
        self._session.targets.sigItemKeyAboutToChange.connect(self._onSessionTargetKeyAboutToChange)
        self._session.targets.sigItemKeyChanged.connect(
            self._onSessionTargetKeyChanged)

        self._currentPoseMetrics = PoseMetricCalculator(
            session=self._session,
            sample=Sample(
                key='CurrentPose',
                timestamp=pd.Timestamp.now()
            )
        )
        self.sigCurrentTargetChanged.connect(lambda: setattr(self._currentPoseMetrics.sample, 'targetKey', self.currentTargetKey))
        self.sigCurrentCoilPositionChanged.connect(self._updateCurrentPoseMetricsSample)
        # TODO: connect signal so that change in activeCoilKey is reflected in currentPoseMetrics.sample.coilKey

        self._currentSamplePoseMetrics = PoseMetricCalculator(
            session=self._session,
            sample=self._session.samples[self._currentSampleKey] if self._currentSampleKey is not None else None
        )

        self.sigCurrentTargetChanged.connect(lambda: self._needToCheckIfOnTarget.set())

        if self._doMonitorOnTarget:
            self._startMonitoringOnTarget()

    @property
    def _cachedActiveCoilKey(self):
        return self.__cachedActiveCoilKey

    @_cachedActiveCoilKey.setter
    def _cachedActiveCoilKey(self, newKey: str | None):
        if newKey == self.__cachedActiveCoilKey:
            return
        oldKey = self.__cachedActiveCoilKey
        logger.info(f'Recording active coil key change from {oldKey} to {newKey}')

        if newKey is not None and self.__cachedActiveCoil is self._session.tools[newKey]:
            # key of existing active coil changed, don't need to disconnect/reconnect signals
            self.__cachedActiveCoilKey = newKey
        else:
            if self.__cachedActiveCoil is not None:
                self.__cachedActiveCoil.sigKeyChanged.disconnect(self._onActiveCoilToolKeyChanged)
                self.__cachedActiveCoil.sigItemChanged.disconnect(self._onActiveCoilToolChanged)
                self.__cachedActiveCoil = None

            self.__cachedActiveCoilKey = newKey
            if newKey is not None:
                self.__cachedActiveCoil = self._session.tools[newKey]
                self.__cachedActiveCoil.sigKeyChanged.connect(self._onActiveCoilToolKeyChanged)
                self.__cachedActiveCoil.sigItemChanged.connect(self._onActiveCoilToolChanged)

        self.sigActiveCoilKeyChanged.emit()

    def _onToolsAboutToChange(self, keys: list[str], attribs: list[str] | None = None):
        if attribs is None or 'isActive' in attribs:
            self._cachedActiveCoilKey = None  # clear cache so that active coil is re-detected on next request

    def _onToolsChanged(self, keys: list[str], attribs: list[str] | None = None):
        if attribs is None or 'isActive' in attribs:
            self._cachedActiveCoilKey = None  # clear cache so that active coil is re-detected on next request

    def _onActiveCoilToolKeyChanged(self, oldKey: str, newKey: str):
        assert oldKey == self._cachedActiveCoilKey
        self._cachedActiveCoilKey = newKey

    def _onActiveCoilToolChanged(self, key: str, attribs: tp.Optional[list[str]] = None):
        """
        This is called whenever any attribute of active coil tool changes, not just when it becomes inactive
        """
        if attribs is None:
            if key in self._session.tools and self._session.tools[key].isActive:
                # tool is still active
                self._currentCoilToMRITransform = None  # clear any previously cached value
                self.sigCurrentCoilPositionChanged.emit()  # transform may have changed
                return
        elif 'isActive' not in attribs:
            assert self._session.tools[key].isActive
            self._currentCoilToMRITransform = None  # clear any previously cached value
            self.sigCurrentCoilPositionChanged.emit()  # transform may have changed
            return

        # tool is no longer active
        self._cachedActiveCoilKey = None

        self._currentCoilToMRITransform = None  # clear any previously cached value
        self.sigCurrentCoilPositionChanged.emit()

    def _onSessionTargetsChanged(self, targetKeysChanged: tp.List[str], targetAttribsChanged: tp.Optional[tp.List[str]]):
        if self._currentTargetKey is not None and self._currentTargetKey in targetKeysChanged:

            if self._currentTargetKey not in self.session.targets:
                logger.debug('Current target deleted')
                self.currentTargetKey = None  # this will emit sigCurrentTargetChanged
                return

            logger.debug('Current target changed')
            self.sigCurrentTargetChanged.emit()

    def _onSessionSamplesChanged(self, sampleKeysChanged: tp.List[str], sampleAttribsChanged: tp.Optional[tp.List[str]]):
        if self._currentSampleKey is not None and self._currentSampleKey in sampleKeysChanged:
            self.sigCurrentSampleChanged.emit()

    def _onLatestPositionsChanged(self):
        self._currentCoilToMRITransform = None  # clear any previously cached value
        self.sigCurrentCoilPositionChanged.emit()
        self.sigCurrentSubjectPositionChanged.emit()
        self._needToCheckIfOnTarget.set()

    async def _loop_monitorOnTarget(self):
        while True:
            await self._needToCheckIfOnTarget.wait()
            self._checkIfOnTarget()
            await asyncio.sleep(1/self._monitorOnTargetRate)

    def _checkIfOnTarget(self):
        if not self._doMonitorOnTarget:
            return

        self._needToCheckIfOnTarget.clear()

        if self._isOnTarget:
            # previously on target
            isStillOnTarget = True
            if isStillOnTarget:
                depthAngleError = self._currentPoseMetrics.getDepthAngleError()
                if np.isnan(depthAngleError) or abs(depthAngleError) > self._isOffTargetWhenZAngleErrorExceeds:
                    isStillOnTarget = False

            if isStillOnTarget:
                horizAngleError = self._currentPoseMetrics.getHorizAngleError()
                if np.isnan(horizAngleError) or abs(horizAngleError) > self._isOffTargetWhenHorizAngleErrorExceeds:
                    isStillOnTarget = False

            if isStillOnTarget:
                horizDistError = self._currentPoseMetrics.getTargetErrorAtCoil()
                if np.isnan(horizDistError) or abs(horizDistError) > self._isOffTargetWhenDistErrorExceeds:
                    isStillOnTarget = False

            if isStillOnTarget:
                depthDistError = self._currentPoseMetrics.getDepthOffsetError()
                if np.isnan(depthDistError) or abs(depthDistError) > self._isOffTargetWhenZDistErrorExceeds:
                    isStillOnTarget = False

            if self._onTargetMaybeChangedAtTime is not None:
                if isStillOnTarget:
                    self._onTargetMaybeChangedAtTime = None
                else:
                    currentTime = time.time()
                    if currentTime - self._onTargetMaybeChangedAtTime > self._isOffTargetMinTime:
                        self._onTargetMaybeChangedAtTime = None
                        self._isOnTarget = False
                        logger.debug(f'isOnTarget: {self._isOnTarget}')
                        self.sigIsOnTargetChanged.emit()

            elif not isStillOnTarget:
                self._onTargetMaybeChangedAtTime = time.time()
                if self._isOffTargetMinTime > 0:
                    pass # don't report change yet, not sure if this is an erroneous measure
                else:
                    self._isOnTarget = False
                    logger.debug(f'isOnTarget: {self._isOnTarget}')
                    self.sigIsOnTargetChanged.emit()

            else:
                pass  # still on target

        else:
            # not previously on target
            isStillOffTarget = False
            if not isStillOffTarget:
                depthAngleError = self._currentPoseMetrics.getDepthAngleError()
                if np.isnan(depthAngleError) or abs(depthAngleError) > self._isOnTargetWhenZAngleErrorUnder:
                    isStillOffTarget = True

            if not isStillOffTarget:
                horizAngleError = self._currentPoseMetrics.getHorizAngleError()
                if np.isnan(horizAngleError) or abs(horizAngleError) > self._isOnTargetWhenHorizAngleErrorUnder:
                    isStillOffTarget = True

            if not isStillOffTarget:
                horizDistError = self._currentPoseMetrics.getTargetErrorAtCoil()
                if np.isnan(horizDistError) or abs(horizDistError) > self._isOnTargetWhenDistErrorUnder:
                    isStillOffTarget = True

            if not isStillOffTarget:
                depthDistError = self._currentPoseMetrics.getDepthOffsetError()
                if (np.isnan(depthDistError) or
                        self._isOnTargetWhenZDistErrorBetween[0] > depthDistError or
                        depthDistError > self._isOnTargetWhenZDistErrorBetween[1]):

                    isStillOffTarget = True

            if self._onTargetMaybeChangedAtTime is not None:
                if isStillOffTarget:
                    self._onTargetMaybeChangedAtTime = None
                else:
                    currentTime = time.time()
                    if currentTime - self._onTargetMaybeChangedAtTime > self._isOnTargetMinTime:
                        self._onTargetMaybeChangedAtTime = None
                        self._isOnTarget = True
                        logger.debug(f'isOnTarget: {self._isOnTarget}')
                        self.sigIsOnTargetChanged.emit()

            elif not isStillOffTarget:
                self._onTargetMaybeChangedAtTime = time.time()
                if self._isOnTargetMinTime > 0:
                    pass  # don't report change yet, wait to be stable on target
                else:
                    self._isOnTarget = True
                    logger.debug(f'isOnTarget: {self._isOnTarget}')
                    self.sigIsOnTargetChanged.emit()

            else:
                pass  # still off target

        if self._onTargetMaybeChangedAtTime is not None:
            self._needToCheckIfOnTarget.set()  # check back again soon, even if tool positions didn't change


    @property
    def session(self):
        return self._session

    @property
    def positionsClient(self):
        return self._positionsClient

    @property
    def currentTargetKey(self):
        return self._currentTargetKey

    @currentTargetKey.setter
    def currentTargetKey(self, newKey: tp.Optional[str]):
        if self._currentTargetKey == newKey:
            return
        self._currentTargetKey = newKey
        logger.debug(f'Current target key changed to {newKey}')
        self.sigCurrentTargetChanged.emit()

    @property
    def currentTarget(self) -> tp.Optional[Target]:
        if self.currentTargetKey is not None:
            return self._session.targets[self.currentTargetKey]
        else:
            return None

    @property
    def currentSampleKey(self):
        return self._currentSampleKey

    @currentSampleKey.setter
    def currentSampleKey(self, newKey: tp.Optional[str]):
        if self._currentSampleKey == newKey:
            return
        self._currentSampleKey = newKey
        self._currentSamplePoseMetrics.sample = self.currentSample
        self.sigCurrentSampleChanged.emit()

    @property
    def currentSample(self) -> tp.Optional[Sample]:
        if self.currentSampleKey is not None:
            return self._session.samples[self.currentSampleKey]
        else:
            return None

    @property
    def currentPoseMetrics(self) -> PoseMetricCalculator:
        return self._currentPoseMetrics

    @property
    def currentSamplePoseMetrics(self) -> PoseMetricCalculator:
        return self._currentSamplePoseMetrics

    @property
    def activeCoilTool(self):
        if self.activeCoilKey is None:
            return None
        else:
            return self._session.tools[self.activeCoilKey]

    @property
    def activeCoilKey(self) -> str | None:
        if self._cachedActiveCoilKey is None:
            # if no active coil specified, use topmost active coil in list
            for toolKey, tool in self._session.tools.items():
                if tool.isActive and isinstance(tool, CoilTool):
                    self._cachedActiveCoilKey = toolKey
                    logger.info(f'Detected active coil key {self._cachedActiveCoilKey}')
                    break
            if self._cachedActiveCoilKey is None:
                return None
        return self._cachedActiveCoilKey

    @property
    def currentCoilToMRITransform(self) -> tp.Optional[Transform]:
        if self._currentCoilToMRITransform is None:
            if self.activeCoilTool is None:
                # no coil active
                return None
            coilTrackerToCameraTransf = self._positionsClient.getLatestTransf(self.activeCoilTool.trackerKey, None)
            subjectTrackerToCameraTransf = self._positionsClient.getLatestTransf(self._session.tools.subjectTracker.trackerKey,
                                                                                 None)
            coilToTrackerTransf = self.activeCoilTool.toolToTrackerTransf

            subjectTrackerToMRITransf = self._session.subjectRegistration.trackerToMRITransf

            if coilToTrackerTransf is None \
                    or coilTrackerToCameraTransf is None \
                    or subjectTrackerToCameraTransf is None\
                    or subjectTrackerToMRITransf is None:
                # cannot compute valid position
                return None

            coilToMRITransform = concatenateTransforms([
                coilToTrackerTransf,
                coilTrackerToCameraTransf,
                invertTransform(subjectTrackerToCameraTransf),
                subjectTrackerToMRITransf
            ])

            self._currentCoilToMRITransform = coilToMRITransform

        return self._currentCoilToMRITransform

    @property
    def doMonitorOnTarget(self):
        return self._doMonitorOnTarget

    @doMonitorOnTarget.setter
    def doMonitorOnTarget(self, doMonitor: bool):
        if doMonitor == self._doMonitorOnTarget:
            return
        self._doMonitorOnTarget = doMonitor
        if doMonitor:
            self._startMonitoringOnTarget()
        else:
            self._stopMonitoringOnTarget()

    @property
    def isOnTarget(self):
        if not self._doMonitorOnTarget:
            self.doMonitorOnTarget = True
        assert self._doMonitorOnTarget
        if self._needToCheckIfOnTarget.is_set():
            # check immediately to make sure we don't give outdated information
            self._checkIfOnTarget()
        return self._isOnTarget

    def _startMonitoringOnTarget(self):
        assert self._monitorOnTargetTask is None
        self._needToCheckIfOnTarget.set()
        self._monitorOnTargetTask = asyncio.create_task(asyncTryAndLogExceptionOnError(self._loop_monitorOnTarget))

    def _stopMonitoringOnTarget(self):
        assert self._monitorOnTargetTask is not None
        self._monitorOnTargetTask.cancel()
        self._monitorOnTargetTask = None

    def _onSessionTargetKeyAboutToChange(self, fromKey: str, toKey: str):
        if self._currentTargetKey == fromKey:
            self._currentTargetKey = toKey
            # other changes will be handled when sigItemsChanged is emitted later

    def _onSessionTargetKeyChanged(self, fromKey: str, toKey: str):
        if fromKey == self._currentPoseMetrics.sample.targetKey:
            self._currentPoseMetrics.sample.targetKey = toKey

    def _updateCurrentPoseMetricsSample(self):
        self._currentPoseMetrics.sample.timestamp = pd.Timestamp.now()
        self._currentPoseMetrics.sample.coilToMRITransf = self.currentCoilToMRITransform

    def createTargetFromCurrentSample(self, doAddToSession: bool = True) -> Target:
        logger.info('Creating target from current sample')
        currentSample = self.currentSample
        if currentSample is None:
            raise ValueError('No sample currently set')
        return self._createTargetFromSample(sample=currentSample, doAddToSession=doAddToSession)

    def createTargetFromCurrentPose(self, doAddToSession: bool = True) -> Target:
        logger.info('Creating target from current pose')
        sample = attrs.evolve(self._currentPoseMetrics.sample,
                              _key='Pose ' + self._currentPoseMetrics.sample.timestamp.strftime('%y.m.%d %H:%M:%S.%f'))
        return self._createTargetFromSample(sample=sample, doAddToSession=doAddToSession)

    def _createTargetFromSample(self, sample: Sample, doAddToSession) -> Target:
        baseKey = sample.key
        targetKey = baseKey
        counter = 1
        while targetKey in self._session.targets:
            counter += 1
            targetKey = f'{baseKey} ({counter})'


        calculator = PoseMetricCalculator(sample=sample, session=self.session)

        coilToScalpDist = calculator.getSampleCoilToScalpDist()
        coilToBrainDist = calculator.getSampleCoilToCortexDist()
        handleAngle = calculator.getAngleFromMidline()

        targetCoord = applyTransform(sample.coilToMRITransf, np.asarray([0, 0, -coilToBrainDist]))
        entryCoord = applyTransform(sample.coilToMRITransf, np.asarray([0, 0, -coilToScalpDist]))

        target = Target(
            key=targetKey,
            targetCoord=targetCoord,
            entryCoord=entryCoord,
            angle=handleAngle,
            depthOffset=coilToScalpDist,
            coilToMRITransf=sample.coilToMRITransf
        )

        if doAddToSession:
            self.session.targets.addItem(target)

        return target

    def getTargetingCoord(self, orientation: str, depth: tp.Union[str, ProjectionSpecification]) -> tp.Optional[np.ndarray]:
        """
        Convenience function for getting a specific coordinate related to targeting orientations.

        Abstracts some of the math needed for things like projecting the current coil orientation down the depth axis
        to plane of target.

        May return None if we are currently missing pose information for a tracker, etc.
        """
        match depth:
            case ProjectionSpecification():
                raise NotImplementedError  # TODO
            case 'coil':
                match orientation:
                    case 'target':
                        if self.currentTarget is None:
                            coilCoord = None
                        else:
                            coilCoord = self.currentTarget.entryCoordPlusDepthOffset
                    case 'coil':
                        transf = self.currentCoilToMRITransform
                        if transf is None:
                            coilCoord = None
                        else:
                            coilCoord = applyTransform(transf, np.asarray([0, 0, 0]), doCheck=False)
                    case _:
                        raise NotImplementedError
                return coilCoord
            case 'skin':
                raise NotImplementedError  # TODO
            case 'gm':
                raise NotImplementedError  # TODO
            case 'target':
                match orientation:
                    case 'target':
                        if self.currentTarget is None:
                            targetCoord = None
                        else:
                            targetCoord = self.currentTarget.targetCoord
                    case 'coil':
                        transf = self.currentCoilToMRITransform
                        if transf is None:
                            targetCoord = None
                        else:
                            if self.currentTarget is None:
                                targetCoord = None
                            else:
                                targetCoord = applyTransform(
                                    transf, np.asarray([0, 0, -np.linalg.norm(
                                        self.currentTarget.entryCoordPlusDepthOffset \
                                        - self.currentTarget.targetCoord)]),
                                    doCheck=False)
                    case _:
                        raise NotImplementedError
                return targetCoord
            case _:
                raise NotImplementedError

    @classmethod
    def getSingleton(cls, session: Session) -> TargetingCoordinator:
        """
        Get the singleton instance of the TargetingCoordinator for the given session.
        """
        global _targetingCoordinatorSingleton
        if _targetingCoordinatorSingleton is None:
            _targetingCoordinatorSingleton = cls(session=session)
        else:
            assert _targetingCoordinatorSingleton.session is session
        return _targetingCoordinatorSingleton

    @staticmethod
    def _resetSingleton():
        global _targetingCoordinatorSingleton
        _targetingCoordinatorSingleton = None




import attrs
import logging

from NaviNIBS.Navigator.GUI.CollectionModels import CollectionTableModel, K
from NaviNIBS.Navigator.Model.Samples import Samples, Sample


logger = logging.getLogger(__name__)


@attrs.define(slots=False)
class SamplesTableModel(CollectionTableModel[str, Samples, Sample]):
    _collection: Samples = attrs.field(init=False, repr=False)
    _isSelectedAttr: str = 'isSelected'

    def __attrs_post_init__(self):
        self._collection = self._session.samples

        self._boolColumns.extend(['isVisible', 'hasTransf'])
        self._editableColumns.extend(['isVisible'])

        self._columns = [
            'key',
            'isVisible',
            'hasTransf',
            'targetKey',
            'timestamp',
            # TODO: add color
        ]
        self._attrColumns = self._columns.copy()
        self._columnLabels = dict(
            key='Key',
            isVisible='Visibility',
            hasTransf='Has pose',
            targetKey='Target',
            timestamp='Time'
        )

        self._collection.sigItemsAboutToChange.connect(self._onSamplesAboutToChange)
        self._collection.sigItemsChanged.connect(self._onSamplesChanged)

        super().__attrs_post_init__()

        # simulate a change to initialize metadata columns
        self._onSamplesAboutToChange(None, None)
        self._onSamplesChanged(None, None)

    def _onSamplesAboutToChange(self, keys: list[str] | None, attrKeys: list[str] | None = None):
        self._onCollectionAboutToChange(keys, attrKeys)

    def _onSamplesChanged(self, keys: list[str] | None, attrKeys: list[str] | None = None):
        self._onCollectionChanged(keys, attrKeys)

        if attrKeys is None or 'metadata' in attrKeys:
            # may have added new metadata field, check if we should add a derived column for it
            sampleKeys = keys
            if sampleKeys is None:
                sampleKeys = self._collection.keys()
            for sampleKey in sampleKeys:
                sample = self._collection[sampleKey]
                for metadataKey in sample.metadata.keys():
                    if metadataKey not in self.derivedColumns:
                        assert metadataKey not in self.columns
                        with self.modifyingColumns():
                            self.derivedColumns[metadataKey] = lambda sampleKey, metadataKey=metadataKey: self._collection[sampleKey].metadata.get(metadataKey, None)
                            self.columns.append(metadataKey)





import asyncio
import logging
import pytest

from NaviNIBS.util.Asyncio import asyncNonCancellable, asyncAtomicCancel

logger = logging.getLogger(__name__)


async def fn1():
    logger.debug('Starting task 1')
    await asyncio.sleep(1.)
    logger.debug('Done task 1')


@asyncNonCancellable
async def fn2():
    logger.debug('Starting task 2')
    await asyncio.sleep(1.)
    logger.debug('Done task 2')


@asyncAtomicCancel
async def fn3():
    logger.debug('Starting task 3')
    await asyncio.sleep(1.)
    logger.debug('Done task 3')


@pytest.mark.asyncio
async def test_cancellation():
    logger.debug('Starting test')
    task1 = asyncio.create_task(fn1())
    task2 = asyncio.create_task(fn2())
    task3 = asyncio.create_task(fn3())
    await asyncio.sleep(0.5)
    logger.debug('Cancelling tasks')
    for task in (task1, task2, task3):
        task.cancel()
    logger.debug('Waiting for tasks to finish')
    with pytest.raises(asyncio.CancelledError):
        await task1  # should have been cancelled

    with pytest.raises(asyncio.CancelledError):
        await task3  # should have finished, but then raised CancelledError

    await task2  # should not have been cancelled
    logger.debug('Done test')




from __future__ import annotations

import shutil

import attrs
from datetime import datetime
import nibabel as nib
import json
import logging
import numpy as np
import os
import pandas as pd
import platformdirs
import pyvista as pv
import tempfile
import typing as tp
from typing import ClassVar

from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.numpy import array_equalish, attrsWithNumpyAsDict, attrsWithNumpyFromDict
from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.Devices import positionsServerHostname, positionsServerPubPort, positionsServerCmdPort

from NaviNIBS.Navigator.Model.GenericCollection import GenericCollection, GenericCollectionDictItem


logger = logging.getLogger(__name__)


SurfMesh = pv.PolyData


@attrs.define
class Tool(GenericCollectionDictItem[str]):
    _usedFor: str  # e.g. 'subject', 'coil', 'pointer'
    _label: tp.Optional[str] = None
    """
    Optional "nice" label for display. If not specified, ``key`` will be used instead. Label does not need to be unique, but could be confusing in some GUI displays if multiple tools with the same label are shown.
    """
    _trackerKey: tp.Optional[str] = None
    """
    Optional key to match up to tracking data. If not specified, ``key`` will be used instead.
    """
    _isActive: bool = True
    """
    Whether to actively use for tracking/pointing
    """
    _doRenderTool: bool = True
    """
    Whether to show tool in camera view, etc. To actually render, a valid pose and mesh must be available. 
    """
    _doRenderTracker: bool = True
    """
    Whether to show tracker in camera view, etc. To actually render, a valid pose and mesh must be available.
    """
    _doShowTrackingState: bool = True
    """
    Whether to include this tool in tracking status widget(s). To actually show, must not also be excluded by other hide filters on the widget.
    """
    _romFilepath: tp.Optional[str] = None
    """
    Path to a .rom file describing optical marker positions on a rigid body tracker, as used by NDI Polaris systems.
    """
    _trackerStlFilepath: tp.Optional[str] = None
    """
    Path to a surface mesh file (e.g. .stl, .ply) for visualization of the tracker.
    """
    _toolStlFilepath: tp.Optional[str] = None
    """
    Path to a surface mesh file (e.g. .stl, .ply) for visualization of the tool.
    """
    _filepathsRelTo: str = '<userDataDir>'
    """
    Can be one of:
    
    - ``'<install>'``: relative to NaviNIBS install dir
    - ``'<userDataDir>'``: relative to user data dir
    - ``'<session>'``: relative to session file
    - absolute path
    """
    _toolToTrackerTransf: tp.Optional[np.ndarray] = None
    """
    Used for aligning actual tool position to Polaris-reported tracker position (e.g. actual coil to coil tracker, or actual pointer to uncalibrated pointer tracker)
    """
    _toolStlToToolTransf: tp.Optional[
        np.ndarray] = None
    """
    Used for visualization of tool surface mesh only; can be used to align mesh with actual tool orientation
    """
    _trackerStlToTrackerTransf: tp.Optional[np.ndarray] = None  # used for visualization of tracker STL only; can be used to align STL with actual reported tracker orientation
    _toolColor: str | None = None
    _trackerColor: str | None = None
    """
    Note: some surf file formats (e.g. .ply) allow specifying color of elements within the file; if color is None here and colors are available in the surf file, those colors will be used.
    """
    _toolOpacity: float | None = None
    _trackerOpacity: float | None = None

    _installPath: tp.Optional[str] = None  # used for relative paths
    _sessionPath: tp.Optional[str] = None  # used for relative paths

    _trackerSurf: tp.Optional[SurfMesh] = attrs.field(init=False, default=None)
    _toolSurf: tp.Optional[SurfMesh] = attrs.field(init=False, default=None)

    _initialTrackerPose: tp.Optional[np.ndarray] = None
    """
    For defining initial pose of tool, e.g. for tools that never get a camera-reported position, or 
    for a default (simulated) position when a camera is not connected. 
    """
    _initialTrackerPoseRelativeTo: str = 'world'

    _toolToTrackerTransfHistory: tp.Dict[str, tp.Optional[np.ndarray]] = attrs.field(factory=dict)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        if self._installPath is None:
            self._installPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', '..')

    @property
    def label(self):
        if self._label is None:
            return self.key
        else:
            return self._label

    @label.setter
    def label(self, newLabel: tp.Optional[str]):
        if self._label == newLabel:
            return
        self.sigItemAboutToChange.emit(self._key, ['label'])
        self._label = newLabel
        self.sigItemChanged.emit(self._key, ['label'])

    @property
    def labelIsSet(self):
        return self._label is not None

    @property
    def trackerKey(self):
        if self._trackerKey is None:
            return self._key
        else:
            return self._trackerKey

    @trackerKey.setter
    def trackerKey(self, newKey: tp.Optional[str]):
        if self._trackerKey == newKey:
            return
        self.sigItemAboutToChange.emit(self._key, ['trackerKey'])
        self._trackerKey = newKey
        self.sigItemChanged.emit(self._key, ['trackerKey'])

    @property
    def trackerKeyIsSet(self):
        return self._trackerKey is not None

    @property
    def usedFor(self):
        return self._usedFor

    @usedFor.setter
    def usedFor(self, newUsedFor: str):
        if self._usedFor == newUsedFor:
            return

        logger.info('Changing {} usedFor from {} to {}'.format(self.key, self._usedFor, newUsedFor))
        self.sigItemAboutToChange.emit(self._key, ['usedFor'])
        self._usedFor = newUsedFor
        self.sigItemChanged.emit(self._key, ['usedFor'])

    @property
    def isActive(self):
        return self._isActive

    @isActive.setter
    def isActive(self, newIsActive: bool):
        if self._isActive == newIsActive:
            return

        logger.info('Changing {} isActive to {}'.format(self.key, newIsActive))
        self.sigItemAboutToChange.emit(self._key, ['isActive'])
        self._isActive = newIsActive
        self.sigItemChanged.emit(self._key, ['isActive'])

    @property
    def doRenderTool(self):
        return self._doRenderTool

    @doRenderTool.setter
    def doRenderTool(self, newDoRenderTool: bool):
        if self._doRenderTool == newDoRenderTool:
            return

        logger.info('Changing {} doRenderTool to {}'.format(self.key, newDoRenderTool))
        self.sigItemAboutToChange.emit(self._key, ['doRenderTool'])
        self._doRenderTool = newDoRenderTool
        self.sigItemChanged.emit(self._key, ['doRenderTool'])

    @property
    def doRenderTracker(self):
        return self._doRenderTracker

    @doRenderTracker.setter
    def doRenderTracker(self, newDoRenderTracker: bool):
        if self._doRenderTracker == newDoRenderTracker:
            return

        logger.info('Changing {} doRenderTracker to {}'.format(self.key, newDoRenderTracker))
        self.sigItemAboutToChange.emit(self._key, ['doRenderTracker'])
        self._doRenderTracker = newDoRenderTracker
        self.sigItemChanged.emit(self._key, ['doRenderTracker'])

    @property
    def doShowTrackingState(self):
        return self._doShowTrackingState

    @doShowTrackingState.setter
    def doShowTrackingState(self, newDoShowTrackingState: bool):
        if self._doShowTrackingState == newDoShowTrackingState:
            return

        logger.info('Changing {} doShowTrackingState to {}'.format(self.key, newDoShowTrackingState))
        self.sigItemAboutToChange.emit(self._key, ['doShowTrackingState'])
        self._doShowTrackingState = newDoShowTrackingState
        self.sigItemChanged.emit(self._key, ['doShowTrackingState'])

    @property
    def romFilepath(self):
        if self._romFilepath is None:
            return None
        else:
            return os.path.join(self.filepathsRelTo, self._romFilepath)

    @romFilepath.setter
    def romFilepath(self, newFilepath: tp.Optional[str]):
        if newFilepath == self.romFilepath:
            return
        logger.info('Changing {} romFilepath to {}'.format(self.key, newFilepath))
        self.sigItemAboutToChange.emit(self.key, ['romFilepath'])
        self._romFilepath = os.path.relpath(newFilepath, self.filepathsRelTo)
        self.sigItemChanged.emit(self.key, ['romFilepath'])

    @property
    def toolStlFilepath(self):
        if self._toolStlFilepath is None:
            return None
        else:
            return os.path.join(self.filepathsRelTo, self._toolStlFilepath)

    @toolStlFilepath.setter
    def toolStlFilepath(self, newFilepath: tp.Optional[str]):
        if newFilepath == self.toolStlFilepath:
            return
        logger.info('Changing {} toolStlFilepath to {}'.format(self.key, newFilepath))
        self._toolSurf = None
        self.sigItemAboutToChange.emit(self.key, ['toolStlFilepath'])
        self._toolStlFilepath = os.path.relpath(newFilepath, self.filepathsRelTo)
        self.sigItemChanged.emit(self.key, ['toolStlFilepath'])

    @property
    def trackerStlFilepath(self):
        if self._trackerStlFilepath is None:
            return None
        else:
            return os.path.join(self.filepathsRelTo, self._trackerStlFilepath)

    @trackerStlFilepath.setter
    def trackerStlFilepath(self, newFilepath: tp.Optional[str]):
        if newFilepath == self.trackerStlFilepath:
            return
        logger.info('Changing {} trackerStlFilepath to {}'.format(self.key, newFilepath))
        self._trackerSurf = None
        self.sigItemAboutToChange.emit(self.key, ['trackerStlFilepath'])
        self._trackerStlFilepath = os.path.relpath(newFilepath, self.filepathsRelTo)
        self.sigItemChanged.emit(self.key, ['trackerStlFilepath'])

    @property
    def filepathsRelTo(self):
        match self._filepathsRelTo:
            case '<install>':
                assert self._installPath is not None
                return self._installPath
            case '<userDataDir>':
                return platformdirs.user_data_dir(appname='NaviNIBS', appauthor=False)
            case '<session>':
                assert self._sessionPath is not None
                return self._sessionPath
            case _:
                return self._filepathsRelTo

    @property
    def filepathsRelToKey(self):
        if self._filepathsRelTo in ('<install>', '<userDataDir>', '<session>'):
            return self._filepathsRelTo
        else:
            return None

    @property
    def sessionPath(self):
        return self._sessionPath

    @sessionPath.setter
    def sessionPath(self, newPath: tp.Optional[str]):
        if self._sessionPath == newPath:
            return

        filepathAttribs = [
            'romFilepath',
            'toolStlFilepath',
            'trackerStlFilepath'
            'sessionPath'
        ]

        if self._filepathsRelTo == '<session>':
            self.sigItemAboutToChange.emit(self.key, filepathAttribs)

        self._sessionPath = newPath

        if self._filepathsRelTo == '<session>':
            self.sigItemChanged.emit(self.key, filepathAttribs)

    @property
    def toolToTrackerTransf(self):
        if self._toolToTrackerTransf is None:
            return np.eye(4)
        else:
            return self._toolToTrackerTransf

    @toolToTrackerTransf.setter
    def toolToTrackerTransf(self, newTransf: tp.Optional[np.ndarray]):
        if array_equalish(self._toolToTrackerTransf, newTransf):
            logger.debug('No change in toolToTrackerTransf, returning')
            return

        # TODO: do validation of newTransf

        self.sigItemAboutToChange.emit(self.key, ['toolToTrackerTransf'])
        logger.info('Set toolToTrackerTransf to {}'.format(newTransf))
        self._toolToTrackerTransf = newTransf
        self._toolToTrackerTransfHistory[self._getTimestampStr()] = None if self._toolToTrackerTransf is None else self._toolToTrackerTransf.copy()
        self.sigItemChanged.emit(self.key, ['toolToTrackerTransf'])

    @property
    def toolStlToToolTransf(self):
        if self._toolStlToToolTransf is None:
            return np.eye(4)
        else:
            return self._toolStlToToolTransf

    @toolStlToToolTransf.setter
    def toolStlToToolTransf(self, newTransf: tp.Optional[np.ndarray]):
        if array_equalish(self._toolStlToToolTransf, newTransf):
            logger.debug('No change in toolStlToToolTransf, returning')
            return

        self.sigItemAboutToChange.emit(self.key, ['toolStlToToolTransf'])
        logger.info('Set toolStlToToolTransf to {}'.format(newTransf))
        self._toolStlToToolTransf = newTransf
        self.sigItemChanged.emit(self.key, ['toolStlToToolTransf'])

    @property
    def trackerStlToTrackerTransf(self):
        if self._trackerStlToTrackerTransf is None:
            return np.eye(4)
        else:
            return self._trackerStlToTrackerTransf

    @trackerStlToTrackerTransf.setter
    def trackerStlToTrackerTransf(self, newTransf: tp.Optional[np.ndarray]):
        if array_equalish(self._trackerStlToTrackerTransf, newTransf):
            logger.debug('No change in trackerStlToTrackerTransf, returning')
            return

        self.sigItemAboutToChange.emit(self.key, ['trackerStlToTrackerTransf'])
        logger.info('Set trackerStlToTrackerTransf to {}'.format(newTransf))
        self._trackerStlToTrackerTransf = newTransf
        self.sigItemChanged.emit(self.key, ['trackerStlToTrackerTransf'])

    @property
    def trackerColor(self):
        return self._trackerColor

    @property
    def toolColor(self):
        return self._toolColor

    @property
    def trackerOpacity(self):
        return self._trackerOpacity

    @property
    def toolOpacity(self):
        return self._toolOpacity

    @property
    def toolToTrackerTransfHistory(self):
        return self._toolToTrackerTransfHistory

    @property
    def trackerSurf(self):
        if self._trackerStlFilepath is not None and self._trackerSurf is None:
            logger.info('Loading tracker mesh from {}'.format(self.trackerStlFilepath))
            self._trackerSurf = pv.read(self.trackerStlFilepath)
        return self._trackerSurf

    @property
    def toolSurf(self):
        if self._toolStlFilepath is not None and self._toolSurf is None:
            logger.info('Loading tool mesh from {}'.format(self.toolStlFilepath))
            self._toolSurf = pv.read(self.toolStlFilepath)
        return self._toolSurf

    @property
    def initialTrackerPose(self):
        return self._initialTrackerPose

    @initialTrackerPose.setter
    def initialTrackerPose(self, newPose: tp.Optional[np.ndarray]):
        if array_equalish(self._initialTrackerPose, newPose):
            logger.debug('No change in initialTrackerPose, returning')
            return

        self.sigItemAboutToChange.emit(self.key, ['initialTrackerPose'])
        logger.info('Set initialTrackerPose to {}'.format(newPose))
        self._initialTrackerPose = newPose
        self.sigItemChanged.emit(self.key, ['initialTrackerPose'])

    @property
    def initialTrackerPoseRelativeTo(self):
        return self._initialTrackerPoseRelativeTo

    @initialTrackerPoseRelativeTo.setter
    def initialTrackerPoseRelativeTo(self, newPoseRelativeTo: tp.Optional[str]):
        if self._initialTrackerPoseRelativeTo == newPoseRelativeTo:
            logger.debug('No change in initialTrackerPoseRelativeTo, returning')
            return

        self.sigItemAboutToChange.emit(self.key, ['initialTrackerPoseRelativeTo'])
        logger.info('Set initialTrackerPoseRelativeTo to {}'.format(newPoseRelativeTo))
        self._initialTrackerPoseRelativeTo = newPoseRelativeTo
        self.sigItemChanged.emit(self.key, ['initialTrackerPoseRelativeTo'])

    def asDict(self) -> tp.Dict[str, tp.Any]:
        npFields = ('toolToTrackerTransf', 'toolStlToToolTransf', 'trackerStlToTrackerTransf', 'initialTrackerPose')
        d = attrsWithNumpyAsDict(self,
                                 npFields=npFields,
                                 eqs=dict(
                                     toolToTrackerTransfHistory=lambda a, b: len(a) == len(b) and
                                                                             ((keyA == keyB and
                                                                               array_equalish(a[keyA], b[keyB]))
                                                                              for keyA, keyB in zip(a, b))))

        if 'toolToTrackerTransfHistory' in d:
            d['toolToTrackerTransfHistory'] = [dict(time=key,
                                                    toolToTrackerTransf=val.tolist() if val is not None else None)
                                               for key, val in d['toolToTrackerTransfHistory'].items()]

        for key in ('installPath', 'sessionPath'):
            d.pop(key, None)

        d['usedFor'] = self.usedFor

        return d

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any], sessionPath: tp.Optional[str] = None):
        if 'toolToTrackerTransfHistory' in d:
            def convertTransf(transf: tp.List[tp.List[float, float, float]]) -> np.ndarray:
                return np.asarray(transf)

            def convertHistoryListToDict(historyList: tp.List[tp.Dict[str, tp.Any]], field: str,
                                         entryConverter: tp.Callable) -> tp.Dict[str, tp.Any]:
                historyDict = {}
                for entry in historyList:
                    timeStr = entry['time']
                    historyDict[timeStr] = entryConverter(entry[field])
                return historyDict

            d['toolToTrackerTransfHistory'] = convertHistoryListToDict(
                d['toolToTrackerTransfHistory'], 'toolToTrackerTransf', convertTransf)

        npFields = ('toolToTrackerTransf', 'toolStlToToolTransf', 'trackerStlToTrackerTransf', 'initialTrackerPose')

        return attrsWithNumpyFromDict(cls, d, npFields=npFields,
                                      sessionPath=sessionPath)

    @staticmethod
    def _getTimestampStr():
        return datetime.today().strftime('%y%m%d%H%M%S.%f')


@attrs.define
class SubjectTracker(Tool):
    _usedFor: str = 'subject'

    @Tool.toolToTrackerTransf.getter
    def toolToTrackerTransf(self):
        """
        Override parent class to assert that transf is identity, since there should never be a
        different transform specified for the subject tracker.
        """
        assert self._toolToTrackerTransf is None
        return np.eye(4)

    @Tool.toolToTrackerTransf.setter
    def toolToTrackerTransf(self, newTransf):
        raise NotImplementedError()  # should never set this transf


@attrs.define
class CoilTool(Tool):
    _usedFor: str = 'coil'

    @Tool.toolToTrackerTransf.getter
    def toolToTrackerTransf(self) -> np.ndarray | None:
        """
        Override parent class to not assume identity transform by default. Unlike other tools, this will
        return None if no transform is set.
        """
        return self._toolToTrackerTransf


@attrs.define
class Pointer(Tool):
    _usedFor: str = 'pointer'


@attrs.define
class CalibrationPlate(Tool):
    _usedFor: str = 'calibration'


@attrs.define
class ToolPositionsServerInfo:
    _hostname: str = positionsServerHostname
    _pubPort: int = positionsServerPubPort
    _cmdPort: int = positionsServerCmdPort
    _type: str = 'IGTLink'
    _doAutostart: bool = True
    _initKwargs: dict[str, tp.Any] = attrs.field(factory=dict)  # args to be passed when starting ToolPositionsServer (e.g. `{'igtlHostname': '192.168.1.123`}`)

    sigInfoChanged: Signal = attrs.field(init=False, factory=lambda: Signal((tp.List[str],)))

    @property
    def hostname(self):
        return self._hostname

    @property
    def pubPort(self):
        return self._pubPort

    @property
    def cmdPort(self):
        return self._cmdPort

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, newType: tp.Optional[str]):
        if newType == self._type:
            return
        self._type = newType
        self.sigInfoChanged.emit(['type'])

    @property
    def doAutostart(self):
        return self._doAutostart

    @property
    def initKwargs(self):
        return self._initKwargs

    def asDict(self) -> tp.Dict[str, tp.Any]:
        return attrsAsDict(self)


@attrs.define
class Tools(GenericCollection[str, Tool]):
    _positionsServerInfo: ToolPositionsServerInfo = attrs.field(factory=ToolPositionsServerInfo)

    _sessionPath: tp.Optional[str] = None  # used for relative paths

    sigPositionsServerInfoChanged: Signal = attrs.field(init=False, factory=lambda: Signal((tp.List[str],)))
    """ includes list of keys of changed info attributes """

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        self._positionsServerInfo.sigInfoChanged.connect(self.sigPositionsServerInfoChanged.emit)

    def addItemFromDict(self, toolDict: dict[str, tp.Any]):
        self.addItem(self.toolFromDict(toolDict, sessionPath=self._sessionPath))

    @property
    def subjectTracker(self) -> tp.Optional[SubjectTracker]:
        subjectTracker = None
        for key, tool in self.items():
            if not tool.isActive:
                continue
            if isinstance(tool, SubjectTracker):
                if subjectTracker is not None:
                    raise ValueError('More than one subject tracker tool is active')
                else:
                    subjectTracker = tool
        return subjectTracker

    @property
    def pointer(self) -> tp.Optional[Pointer]:
        pointer = None
        for key, tool in self.items():
            if not tool.isActive:
                continue
            if isinstance(tool, Pointer):
                if pointer is not None:
                    raise ValueError('More than one pointer tool is active')
                else:
                    pointer = tool
        return pointer

    @property
    def calibrationPlate(self) -> tp.Optional[CalibrationPlate]:
        calibrationPlate = None
        for key, tool in self.items():
            if not tool.isActive:
                continue
            if isinstance(tool, CalibrationPlate):
                if calibrationPlate is not None:
                    raise ValueError('More than one calibration plate is active')
                else:
                    calibrationPlate = tool
        return calibrationPlate

    @property
    def positionsServerInfo(self):
        return self._positionsServerInfo

    def _getActiveToolKeys(self) -> dict[str, tp.Union[str, list[str,...]]]:
        activeToolKeys = {}
        for key, tool in self._tools.items():
            if not tool.isActive:
                continue
            if tool.usedFor in ('subject', 'pointer'):
                assert tool.usedFor not in activeToolKeys, 'More than one active tool for {}'.format(tool.usedFor)

        raise NotImplementedError()  # TODO

    def _checkActiveTools(self):
        raise NotImplementedError()  # TODO: assert that only one subject, pointer, and (for now) coil tracker are active

    @property
    def sessionPath(self):
        return self._sessionPath

    @sessionPath.setter
    def sessionPath(self, newPath: tp.Optional[str]):
        if self._sessionPath == newPath:
            return

        changingKeys = [tool.key for tool in self.values() if tool.filepathsRelTo == '<session>']
        self.sigItemsAboutToChange.emit(changingKeys)
        self._sessionPath = newPath
        with self.sigItemsAboutToChange.blocked(), self.sigItemsChanged.blocked():
            for tool in self.values():
                tool.sessionPath = self._sessionPath
        self.sigItemsChanged.emit(changingKeys)

    def asList(self) -> tp.List[tp.Dict[str, tp.Any]]:
        toolList = super().asList()

        # add ToolPositionsServer info as another entry in list next to tools
        serverInfo = self._positionsServerInfo.asDict()
        if len(serverInfo) > 0:
            serverInfo['key'] = 'ToolPositionsServer'
            toolList.append(serverInfo)

        return toolList

    @classmethod
    def fromList(cls, toolList: tp.List[tp.Dict[str, tp.Any]], sessionPath: tp.Optional[str] = None) -> Tools:

        tools = {}
        serverInfo = None
        for toolDict in toolList:
            key = toolDict['key']
            if key == 'ToolPositionsServer':
                # this is not actually a tool dict, it is connection info for ToolPositionsServer
                serverInfo = ToolPositionsServerInfo(**{key: val for key, val in toolDict.items() if key != 'key'})
            else:
                tools[key] = cls.toolFromDict(toolDict, sessionPath=sessionPath)

        if serverInfo is None:
            serverInfo = ToolPositionsServerInfo()

        return cls(items=tools, sessionPath=sessionPath, positionsServerInfo=serverInfo)

    def _onItemChanged(self, key: str, attribKeys: tp.Optional[list[str]] = None):
        super()._onItemChanged(key=key, attribKeys=attribKeys)
        if attribKeys is not None and 'usedFor' in attribKeys:
            self._onToolUsedForChanged(key=key)

    def _onToolUsedForChanged(self, key: str):
        toolDict = self[key].asDict()
        tool = self.toolFromDict(toolDict, sessionPath=self._sessionPath)
        self.setItem(tool)

    @classmethod
    def toolFromDict(cls, toolDict: tp.Dict[str, tp.Any], sessionPath: tp.Optional[str] = None) -> Tool:
        usedFor = toolDict['usedFor']

        match usedFor:
            case 'coil':
                ToolCls = CoilTool
            case 'pointer':
                ToolCls = Pointer
            case 'calibration':
                ToolCls = CalibrationPlate
            case 'subject':
                ToolCls = SubjectTracker
            case '' | 'visualization':
                ToolCls = Tool
            case _:
                raise NotImplementedError('Unexpected tool usedFor: {}'.format(usedFor))

        tool = ToolCls.fromDict(toolDict, sessionPath=sessionPath)

        return tool
    
    
    


from __future__ import annotations
import asyncio
import attrs
from contextlib import contextmanager
import logging
import numpy as np
import pyvista as pv
from pyvista.plotting.mapper import DataSetMapper
import pyvistaqt as pvqt
from qtpy import QtGui, QtCore, QtWidgets
import typing as tp
from typing import ClassVar

from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.pyvista import Actor, setActorUserTransform
from NaviNIBS.util.pyvista.ticks import getNiceTicksForRange

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class _DelayedPlotter:
    _needsRender: asyncio.Event
    _renderTask: asyncio.Task
    _renderingNotPaused: asyncio.Event
    _pauseStackCount_: int
    minRenderPeriod: float

    def __init__(self, minRenderPeriod: float = 0.05):
        self._needsRender = asyncio.Event()
        self._renderingNotPaused = asyncio.Event()
        self._renderingNotPaused.set()

        self._pauseStackCount_ = 0

        self.minRenderPeriod = minRenderPeriod

        self._renderTask = asyncio.create_task(asyncTryAndLogExceptionOnError(self._renderLoop))

    @property
    def _pauseStackCount(self):
        return self._pauseStackCount_

    @_pauseStackCount.setter
    def _pauseStackCount(self, value: int):
        if value < 0:
            value = 0
        self._pauseStackCount_ = value
        if self._pauseStackCount_ == 0:
            self._renderingNotPaused.set()
        else:
            self._renderingNotPaused.clear()

    def pauseRendering(self):
        self._pauseStackCount += 1

    def maybeResumeRendering(self):
        self._pauseStackCount -= 1

    def resumeRendering(self):
        self._pauseStackCount = 0

    @contextmanager
    def renderingPaused(self):
        self.pauseRendering()
        yield
        self.maybeResumeRendering()

    def _renderNow(self):
        self._needsRender.clear()
        logger.debug('Rendering')
        super().render()
        logger.debug('Done rendering')

    async def _renderLoop(self):
        while True:
            await asyncio.sleep(self.minRenderPeriod)
            await self._needsRender.wait()
            await self._renderingNotPaused.wait()
            self._renderNow()

    def render(self, doRenderImmediately: bool = False):
        if doRenderImmediately:
            logger.debug('Rendering immediately')
            self._renderNow()
            logger.debug('Done rendering immediately')
        else:
            logger.debug('Setting render flag')
            self._needsRender.set()


class PlotterImprovementsMixin:
    def __init__(self):
        pass

    def updateScalarBarRangeWithVol(self, clim: tuple[float, float], volumeKey: str, volume: pv.UniformGrid):
        """
        Handle manual update of volume scalars that were hard-converted to previous scale within add_volume
        """
        scalars = self.mapper.GetInput()
        scalars[volumeKey] = volume[volumeKey].copy().astype(np.float64)
        scalars[volumeKey].clip(clim[0], clim[1], out=scalars[volumeKey])
        scalars[volumeKey] = (scalars[volumeKey] - clim[0]) / (clim[1] - clim[0]) * 255
        self.update_scalar_bar_range(clim)

    def reset_scalar_bar_ranges(self, scalarBarTitles: tp.Optional[list[str]] = None):
        """
        Fix some issues with update_scalar_bar_ranges and handling of all nan values
        """
        if scalarBarTitles is None:
            scalarBarTitles = self.scalar_bars.keys()
        for scalarBarKey in scalarBarTitles:
            clims = [np.nan, np.nan]

            for actorKey, actor in self.renderer.actors.items():
                mapper = actor.GetMapper()
                if mapper in self.scalar_bars._scalar_bar_mappers[scalarBarKey]:
                    # note: this works for actors created by add_mesh, but maybe not others
                    mesh = mapper.GetInput()
                    thisScalarKey = mapper.GetArrayName()
                    scalars = pv.get_array(mesh, thisScalarKey)
                    if not isinstance(scalars, np.ndarray):
                        scalars = np.asarray(scalars)
                    clims = [np.nanmin([clims[0], np.nanmin(scalars)]),
                             np.nanmax([clims[1], np.nanmax(scalars)])]
            if np.isnan(clims[0]):
                clims = [0, 1]
            for mapper in self._scalar_bars._scalar_bar_mappers[scalarBarKey]:
                mapper.scalar_range = clims

    def addLineSegments(self, lines: pv.PolyData,
        color = 'w',
        opacity = 1.,
        width = 5,
        reset_camera: bool = False,
        label = None,
        name = None,
        scalar_bar_args: dict | None = None,
        scalars: str | None = None,
        show_scalar_bar: bool | None = None,
        cmap: str | tp.Any | None = None,
        userTransform: np.ndarray | None = None) -> pv.Actor:
        """
        Similar to plotter.add_lines but with a few improvements:
        - Allows `lines` arg to already be pv.PolyData (e.g. to already been passed through `pv.lines_from_points`.) This
            allows for the possibility of grouping multiple discontinuous line segments into one actor.
        - Added opacity arg
        - Added reset_camera arg
        - Allows specifying user transform in same call
        """
        assert isinstance(lines, pv.PolyData)

        # Create mapper and add lines
        mapper = DataSetMapper(lines)

        if scalars is None:
            assert scalar_bar_args is None
            rgb_color = pv.colors.Color(color)
        else:
            if show_scalar_bar is None:
                show_scalar_bar = self._theme.show_scalar_bar

            n_colors = 256

            scalar_bar_args = scalar_bar_args.copy()
            scalar_bar_args.setdefault('n_colors', n_colors)
            scalar_bar_args['mapper'] = mapper

            nan_color = pv.Color(None, default_opacity=1., default_color=self._theme.nan_color)

            if isinstance(scalars, str):
                scalars = pv.get_array(lines, scalars, preference='points')

            clim = [np.nanmin(scalars), np.nanmax(scalars)]

            if any(np.isnan(x) for x in clim):
                # autoset clim within set_scalars errors if all values are nan
                clim = [0, 1]

            mapper.set_scalars(lines,
                               scalars=scalars,
                               scalar_bar_args=scalar_bar_args,
                               rgb=False,
                               component=None,
                               preference='points',
                               interpolate_before_map=True,
                               _custom_opac=False,
                               annotations=None,
                               log_scale=False,
                               nan_color=nan_color,
                               above_color=None,
                               below_color=None,
                               cmap=cmap,
                               flip_scalars=False,
                               opacity=opacity,
                               categories=False,
                               n_colors=n_colors,
                               clim=clim,
                               theme=self._theme,
                               show_scalar_bar=show_scalar_bar)

        # Create actor
        actor = pv._vtk.vtkActor()
        if userTransform is not None:
            setActorUserTransform(actor, userTransform)
        actor.SetMapper(mapper)
        actor.GetProperty().SetLineWidth(width)
        actor.GetProperty().EdgeVisibilityOn()
        if scalars is None:
            actor.GetProperty().SetEdgeColor(rgb_color.float_rgb)
            actor.GetProperty().SetColor(rgb_color.float_rgb)
        actor.GetProperty().LightingOff()
        actor.GetProperty().SetOpacity(opacity)

        # legend label
        if label:
            if not isinstance(label, str):
                raise TypeError('Label must be a string')
            addr = actor.GetAddressAsString("")
            self.renderer._labels[addr] = [lines, label, rgb_color]

        if show_scalar_bar and scalars is not None:
            self.add_scalar_bar(**scalar_bar_args)

        # Add to renderer
        self.add_actor(actor, reset_camera=reset_camera, name=name, pickable=False)
        self.render()
        return actor

    def add_axes_marker(self,
                           total_length: tuple[float, float, float] | None = None,
                           reset_camera: bool = False,
                           **kwargs) -> Actor:
        actor = pv.create_axes_marker(**kwargs)
        if total_length is not None:
            actor.SetTotalLength(*total_length)

        self.add_actor(actor, reset_camera=reset_camera, pickable=False)

        return actor

    def addMesh(self, mesh, defaultMeshColor: str | None = None, **kwargs):
        """
        Wrapper around add_mesh with some extra defaults for things like mesh color handling

        defaultMeshColor: use this if the mesh has no color arrays
        """
        color = kwargs.pop('color', None)
        scalars = kwargs.pop('scalars', None)
        rgb = kwargs.pop('rgb', None)
        if color is None and scalars is None:
            for arrayName in mesh.array_names:
                if mesh[arrayName].shape[1] in (3, 4):
                    scalars = arrayName
                    break

        if color is None and scalars is None:
            color = defaultMeshColor  # default color if nothing else provided

        if rgb is None:
            rgb = color is None

        try:
            toReturn = self.add_mesh(mesh=mesh,
                                              color=color,
                                              scalars=scalars,
                                              rgb=rgb,
                                              **kwargs)
        except (AttributeError, ValueError) as e:
            from NaviNIBS.util import exceptionToStr
            logger.error(exceptionToStr(e))
            raise e

        return toReturn

    def showGrid(self,
                 mesh: pv.PolyData | None = None,
                 bounds: tp.Sequence[float] | None = None,
                 n_xlabels=5,
                 n_ylabels=5,
                 n_zlabels=5,
                 **kwargs):
        """
        Similar to show_grid, but simplified and with fixes for tick rounding issues
        introduced in https://github.com/pyvista/pyvista/commit/344f8388841043e0ff68b5813c4560f55333f057.

        As part of this, specified number of tick marks will be used as a suggestion, but actual number may be different.
        """
        r: pv.Renderer = self.renderer

        if mesh is None and bounds is None:
            # Use the bounds of all data in the rendering window
            bounds = np.array(r.bounds)
        elif bounds is None:
            # otherwise, use the bounds of the mesh (if available)
            bounds = np.array(mesh.bounds)
        else:
            bounds = np.asanyarray(bounds, dtype=float)

        # round bounds so that tick marks are on nicer intervals
        ticksPerDim = np.array([n_xlabels, n_ylabels, n_zlabels])
        # if aspect ratio is far from 1, then adjust num ticks for narrow dimensions
        widths = bounds[1::2] - bounds[::2]
        for i in range(3):
            while widths[i] / widths.max() / ticksPerDim[i] * ticksPerDim[widths.argmax()] < 0.3:
                if ticksPerDim[i] == 2:
                    break
                ticksPerDim[i] -= 1

        for i in range(3):
            niceTicks = getNiceTicksForRange(bounds[2*i], bounds[2*i+1], ticksPerDim[i])
            bounds[2*i] = niceTicks[0]
            bounds[2*i+1] = niceTicks[-1]
            ticksPerDim[i] = len(niceTicks)

        n_xlabels = ticksPerDim[0]
        n_ylabels = ticksPerDim[1]
        n_zlabels = ticksPerDim[2]

        return self.show_grid(mesh=mesh,
                              bounds=bounds,
                              n_xlabels=n_xlabels,
                              n_ylabels=n_ylabels,
                              n_zlabels=n_zlabels,
                              **kwargs)


class BackgroundPlotter(_DelayedPlotter, pvqt.plotting.QtInteractor, PlotterImprovementsMixin):
    """
    Similar to pvqt.BackgroundPlotter, with a few key differences:
    - This batches multiple render calls together with an async coroutine
    - This sets default background color based on app palette
    - This doesn't include some of the "extra" functionality in pvqt.BackgroundPlotter related to
        toolbars and single-plotter-per-window management.
    """

    def __init__(self, *args,
                 auto_update: float = 0.0,
                 **kwargs):
        _DelayedPlotter.__init__(self, **{key: val for key, val in kwargs.items() if key in ('minRenderPeriod',)})
        try:
            kwargs.pop('minRenderPeriod')
        except KeyError:
            pass

        if True:  # TODO: delete once no callers provide these kwargs
            # for legacy compatibility, drop any kwargs expected by pvqt.BackgroundPlotter but not by pvqt.plotting.QtInteractor
            for key in ('app', 'show'):
                try:
                    kwargs.pop(key)
                except KeyError:
                    pass

        pvqt.plotting.QtInteractor.__init__(self, *args,
                                                 auto_update=auto_update,
                                                 **kwargs)

        #self.enable_anti_aliasing(aa_type='msaa')  # for nice visuals

        self.set_background(self.palette().color(QtGui.QPalette.Base).name())

        if False:  # TODO: debug, disable
            self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.ActionsContextMenu)
            action = QtWidgets.QAction('Export scene to obj', self)
            self.addAction(action)
            action.triggered.connect(self._onExportToObj)

        PlotterImprovementsMixin.__init__(self)

    @contextmanager
    def allowNonblockingCalls(self):
        """
        For now, this is does nothing. Is here for interface compatibility with RemotePlotterProxy that may be used instead.
        :return:
        """
        yield

    @contextmanager
    def disallowNonblockingCalls(self):
        """
        For now, this is does nothing. Is here for interface compatibility with RemotePlotterProxy that may be used instead.
        :return:
        """
        yield

    async def add_lines_async(self, *args, **kwargs):
        """
        Not actually async. Is here for interface compatibility with RemotePlotterProxy that may be used instead
        """
        return self.add_lines(*args, **kwargs)

    async def add_points_async(self, *args, **kwargs):
        """
        Not actually async. Is here for interface compatibility with RemotePlotterProxy that may be used instead
        """
        return self.add_points(*args, **kwargs)

    def _onExportToObj(self):
        exportFilepath, _ = QtWidgets.QFileDialog.getSaveFileName(self,
                                                               'Export scene to obj',
                                                               '',
                                                               'obj (*.obj)')
        if len(exportFilepath) == 0:
            logger.info('Export cancelled')
            return

        self.export_obj(exportFilepath)

    def closeEvent(self, evt):
        toRet = super().closeEvent(evt)
        self.close()
        self.Finalize()  # suggested by https://discourse.vtk.org/t/wglmakecurrent-failed-in-makecurrent-after-closed-a-window-with-two-vtk-widget/5899/2
        return toRet

    def addIrenStyleClassObserver(self, *, event: str, callback: tp.Callable):
        # shortcut alias for self.iren._style_class.AddObserver, with kwargs instead of args, and
        # not passing first arg when calling callback
        # (all for easier implementation with remote plotters)
        return self.iren._style_class.AddObserver(event, lambda _, event: callback(None, event))


class SecondaryLayeredPlotter(_DelayedPlotter, pv.BasePlotter, PlotterImprovementsMixin):
    _mainPlotter: PrimaryLayeredPlotter
    _rendererLayer: int

    def __init__(self, mainPlotter: PrimaryLayeredPlotter, rendererLayer: tp.Optional[int] = None, **kwargs):
        _DelayedPlotter.__init__(self, **{key: val for key, val in kwargs if key in ('minRenderPeriod',)})
        try:
            kwargs.pop('minRenderPeriod')
        except KeyError:
            pass

        pv.BasePlotter.__init__(self, **kwargs)

        if rendererLayer is None:
            rendererLayer = mainPlotter.ren_win.GetNumberOfLayers()
            logger.info('Renderer layer: {rendererLayer}')

        self._rendererLayer = rendererLayer

        self._mainPlotter = mainPlotter

        mainPlotter.ren_win.SetNumberOfLayers(mainPlotter.ren_win.GetNumberOfLayers()+1)
        for renderer in self.renderers:
            mainPlotter.ren_win.AddRenderer(renderer)
            renderer.SetLayer(rendererLayer)
        mainPlotter.link_views_across_plotters(self)

        self.renderer.set_background(mainPlotter.background_color)

        for renderer in self.renderers:
            renderer.SetLayer(rendererLayer)

        self.iren = None

        PlotterImprovementsMixin.__init__(self)

    @contextmanager
    def allowNonblockingCalls(self):
        """
        For now, this is does nothing. Is here for interface compatibility with RemotePlotterProxy that may be used instead.
        :return:
        """
        yield

    @contextmanager
    def disallowNonblockingCalls(self):
        """
        For now, this is does nothing. Is here for interface compatibility with RemotePlotterProxy that may be used instead.
        :return:
        """
        yield

    @property
    def rendererLayer(self):
        return self._rendererLayer

    def enable_depth_peeling(self, *args, **kwargs):
        result = self.renderer.enable_depth_peeling(*args, **kwargs)
        if result:
            self._mainPlotter.ren_win.AlphaBitPlanesOn()
        return result

    def render(self, doRenderPrimaryPlotter: bool = True, doRenderImmediately: bool = False):
        super().render(doRenderImmediately=doRenderImmediately)
        if doRenderPrimaryPlotter:
            self._mainPlotter.render(doRenderSecondaryPlotters=False,
                                     doRenderImmediately=doRenderImmediately)  # TODO: determine if this is necessary

    def reset_camera_clipping_range(self, doIncludeMainPlotter: bool = True):
        if doIncludeMainPlotter:
            self._mainPlotter.reset_camera_clipping_range()
        else:
            super().reset_camera_clipping_range()


from pyvista.plotting.render_window_interactor import RenderWindowInteractor


class CustomRenderWindowInteractor(RenderWindowInteractor):
    """
    Some quirks for layered plotter setup require us to override some behavior in pyvista's RenderWindowInteractor
    """

    def get_event_subplot_loc(self):
        # fix picking behavior for layered plotters

        # assume no subplots
        return 0, 0


class PrimaryLayeredPlotter(BackgroundPlotter):

    _secondaryPlotters: dict[str, SecondaryLayeredPlotter]
    _doAutoAdjustCameraClippingRange: bool

    def __init__(self, *args, rendererLayer: int = 0, doAutoAdjustCameraClippingRange: bool = True, **kwargs):
        self._secondaryPlotters = dict()

        BackgroundPlotter.__init__(self, *args, **kwargs)

        for renderer in self.renderers:
            renderer.SetLayer(rendererLayer)

        self._doAutoAdjustCameraClippingRange = doAutoAdjustCameraClippingRange

        # disable auto-adjust of camera clipping since it fails with multiple renderers
        self.iren.interactor.GetInteractorStyle().AutoAdjustCameraClippingRangeOff()

    @property
    def secondaryPlotters(self):
        return self._secondaryPlotters

    @property
    def doAutoAdjustCameraClippingRange(self):
        return self._doAutoAdjustCameraClippingRange

    @doAutoAdjustCameraClippingRange.setter
    def doAutoAdjustCameraClippingRange(self, value: bool):
        if value == self._doAutoAdjustCameraClippingRange:
            return
        self._doAutoAdjustCameraClippingRange = value
        if value:
            self.render()

    def _setup_interactor(self, off_screen: bool) -> None:
        # override superclass to use custom RenderWindowInteractor
        if off_screen:
            self.iren: tp.Any = None
        else:
            self.iren = CustomRenderWindowInteractor(
                self, interactor=self.ren_win.GetInteractor()
            )
            self.iren.interactor.RemoveObservers(
                "MouseMoveEvent"
            )  # slows window update?
            self.iren.initialize()
            self.enable_trackball_style()

    def setLayer(self, layer: int):
        for renderer in self.renderers:
            renderer.SetLayer(layer)
        # TODO: do validation to make sure ren_win.NumberLayers etc. is set as needed

    def addLayeredPlotter(self, key: str, layer: tp.Optional[int] = None) -> SecondaryLayeredPlotter:
        assert key not in self._secondaryPlotters
        self._secondaryPlotters[key] = SecondaryLayeredPlotter(mainPlotter=self, rendererLayer=layer)
        return self._secondaryPlotters[key]

    def pauseRendering(self):
        super().pauseRendering()
        for plotter in self._secondaryPlotters.values():
            plotter.pauseRendering()

    def resumeRendering(self):
        super().resumeRendering()
        for plotter in self._secondaryPlotters.values():
            plotter.resumeRendering()

    def _renderNow(self):
        if self._doAutoAdjustCameraClippingRange:
            logger.debug('Setting camera clipping range')
            self.reset_camera_clipping_range()
        super()._renderNow()

    def render(self, doRenderSecondaryPlotters: bool = True, doRenderImmediately: bool = False):
        super().render(doRenderImmediately=doRenderImmediately)
        if doRenderSecondaryPlotters:
            for plotter in self._secondaryPlotters.values():
                plotter.render(doRenderPrimaryPlotter=False, doRenderImmediately=doRenderImmediately)

    def set_background(self, *args, **kwargs):
        super().set_background(*args, **kwargs)
        for plotter in self._secondaryPlotters.values():
            plotter.set_background(*args, **kwargs)

    def enable_depth_peeling(self, *args, **kwargs):
        super().enable_depth_peeling(*args, **kwargs)
        # NOTE: apply this to primary plotter only; if want to enable depth peeling in
        #  secondary plotters, then need to call enable_depth_peeling for each separately

    def reset_camera(self):
        bounds = self._getSharedBounds()
        super().reset_camera(bounds=bounds)
        for plotter in self._secondaryPlotters.values():
            plotter.reset_camera(bounds=bounds)

    def _getSharedBounds(self) -> list[float]:
        # manually calculate bounds collectively from primary and secondary plotters
        bounds = np.asarray(self.bounds).reshape((3, 2)).T
        for plotter in self.secondaryPlotters.values():
            thisBounds = np.asarray(plotter.bounds).reshape((3, 2)).T
            bounds[0, :] = np.minimum(bounds[0, :], thisBounds[0, :])
            bounds[1, :] = np.maximum(bounds[1, :], thisBounds[1, :])

        bounds[0, :] = bounds[0, :] - (bounds[1, :] - bounds[0, :])
        bounds[1, :] = bounds[1, :] + (bounds[1, :] - bounds[0, :])

        return bounds.T.reshape((6,)).tolist()

    def reset_camera_clipping_range(self):
        if False:
            super().reset_camera_clipping_range()
            for plotter in self.secondaryPlotters.values():
                plotter.reset_camera_clipping_range(doIncludeMainPlotter=False)
        else:
            bounds = self._getSharedBounds()

            self.renderer.ResetCameraClippingRange(*bounds)

            # for plotter in self.secondaryPlotters.values():
            #     plotter.renderer.ResetCameraClippingRange(*bounds.T.reshape((6,)).tolist())

            logger.debug(f'Camera clipping range: {self.renderer.camera.clipping_range}')

    def set_camera_clipping_range(self, range):
        # TODO: verify whether this is actually needed
        super().camera.clipping_range = range
        for plotter in self.secondaryPlotters.values():
            plotter.camera.clipping_range = range

    def close(self):
        for plotter in self.secondaryPlotters.values():
            plotter.close()
        super().close()


from __future__ import annotations

import asyncio

import attrs
import logging
import numpy as np
import pyvista as pv
import typing as tp
from typing import ClassVar

from . import PlotViewLayer
from NaviNIBS.util.pyvista import Actor, setActorUserTransform, concatenateLineSegments
from NaviNIBS.util.Transforms import applyTransform, concatenateTransforms, invertTransform, composeTransform


logger = logging.getLogger(__name__)


Transform = np.ndarray


@attrs.define
class PlotLayersGroup(PlotViewLayer):
    _type: ClassVar[str] = 'Group'
    _layers: tp.Dict[str, PlotViewLayer] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def addLayer(self, cls: tp.Callable[..., PlotViewLayer], key: str, **kwargs):
        assert key not in self._layers
        self._layers[key] = cls(key=key, **kwargs,
                                coordinator=self._coordinator,
                                plotter=self._plotter,
                                plotInSpace=self._plotInSpace)

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None):
        if not isinstance(which, str):
            # assume parent call above triggered appropriate redraws
            return

        if which == 'all':
            which = 'layers'
            self._redraw(which=which)
            return

        if which == 'layers':
            for layer in self._layers.values():
                layer._redraw()

        else:
            raise NotImplementedError




import asyncio
import attrs
import logging
from qtpy import QtWidgets, QtGui, QtCore
import typing as tp

from NaviNIBS.Devices.ToolPositionsClient import ToolPositionsClient
from NaviNIBS.Navigator.Model.Session import Session, Tool
from NaviNIBS.util.Signaler import Signal

logger = logging.getLogger(__name__)


@attrs.define
class ToolCalibrationWindow:
    """
    Base class for more specific coil or pointer calibration classes.
    """
    _parent: QtWidgets.QWidget
    _toolKeyToCalibrate: str
    _session: Session = attrs.field(repr=False)

    _positionsClient: ToolPositionsClient = attrs.field(init=False)

    _wdgt: QtWidgets.QDialog = attrs.field(init=False)

    sigFinished: Signal = attrs.field(init=False, factory=lambda: Signal((bool,)))  # includes False if calibration was cancelled, else True

    def __attrs_post_init__(self):
        self._positionsClient = ToolPositionsClient()
        self._positionsClient.sigLatestPositionsChanged.connect(self._onLatestPositionsChanged)

        self._wdgt = QtWidgets.QDialog(self._parent)
        self._wdgt.setModal(True)

        self._wdgt.setWindowTitle('{} calibration'.format(self._toolKeyToCalibrate))

        self._wdgt.setWindowModality(QtGui.Qt.WindowModal)

        self._wdgt.finished.connect(self._onDlgFinished)

    def _onDlgFinished(self, result: int):
        self._positionsClient.stopReceivingPositions()
        self.sigFinished.emit(result == QtWidgets.QDialog.Accepted)

    @property
    def toolToCalibrate(self) -> Tool:
        return self._session.tools[self._toolKeyToCalibrate]

    def show(self):
        self._wdgt.show()

    @property
    def wdgt(self):
        return self._wdgt

    @property
    def session(self):
        return self._session

    def _onLatestPositionsChanged(self):
        pass


from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from NaviNIBS.Navigator.Model.Session import Session


def getLogFilepath(session: Session) -> str:
    return os.path.abspath(os.path.join(session.unpackedSessionDir, 'NaviNIBS_Log.txt'))


def createLogFileHandler(logFilepath: str, use_async: bool = True) -> logging.FileHandler:
    """
    Create a handler for logging to a file, using concurrent_log_handler
    to support multiple processes writing to the same log file.
    """
    from concurrent_log_handler import ConcurrentTimedRotatingFileHandler  # noqa

    handler = ConcurrentTimedRotatingFileHandler(
        logFilepath,
        when='D', # TODO: debug, set back to 'D'
        interval=1,
        maxBytes=10**8,
    )
    handler.setFormatter(logging.Formatter(
            fmt='%(asctime)s.%(msecs)03d  %(process)6d %(filename)20s %(lineno)4d %(levelname)5s: %(message)s',
            datefmt='%H:%M:%S'))

    handler.setLevel(logging.DEBUG)

    if use_async:
        # To enable background logging queue, call this near the end of your logging setup.
        from concurrent_log_handler.queue import setup_logging_queues

        setup_logging_queues()

    return handler








from __future__ import annotations

import asyncio

import attrs
import logging
from math import nan, isnan
import numpy as np
import typing as tp
from typing import ClassVar
import pyvista as pv

from NaviNIBS.Navigator.GUI.ViewPanels.NavigatePanel.ViewLayers.OrientationsLayers import SampleOrientationsLayer, VisualizedOrientation
from NaviNIBS.Navigator.GUI.ViewPanels.NavigatePanel.ViewLayers.MeshSurfaceLayer import HeadMeshSurfaceLayer
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.Transforms import applyTransform, invertTransform, composeTransform, concatenateTransforms

logger = logging.getLogger(__name__)


@attrs.define(kw_only=True)
class SampleMetadataOrientationsLayer(SampleOrientationsLayer):

    _type: ClassVar[str] = 'SampleMetadataOrientations'

    _metadataKey: str
    _colorbarLabel: str | None = None
    _metadataScaleFactor: float = 1.0

    _colorDepthIndicator: str | None = None
    _colorHandleIndicator: str | None = None
    _colorDepthIndicatorSelected: str | None = None
    _colorHandleIndicatorSelected: str | None = None

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def _createVisualizedOrientationForSample(self, key: str) -> VisualizedOrientation:

        isSelected = self.orientations[key].isSelected

        metadataVal = nan
        if self._metadataKey in self._coordinator.session.samples[key].metadata:
            metadataVal = self._coordinator.session.samples[key].metadata[self._metadataKey]

        logger.debug(f'metadataVal: {metadataVal}')

        if isnan(metadataVal):
            color = 'gray'
        else:
            if self._colorbarLabel is None:
                colorbarLabel = self._metadataKey
            else:
                colorbarLabel = self._colorbarLabel
            color = (metadataVal * self._metadataScaleFactor, colorbarLabel)

        lineWidth = self._lineWidth
        if isSelected:
            lineWidth *= 2  # increase line width to highlight selected samples

        return VisualizedOrientation(
            orientation=self.orientations[key],
            plotter=self._plotter,
            colorHandleIndicator=color,
            colorDepthIndicator=color,
            opacity=self._opacity,
            lineWidth=lineWidth,
            style=self._style,
            actorKeyPrefix=self._getActorKey(key)
        )


@attrs.define(kw_only=True)
class SampleMetadataInterpolatedSurfaceLayer(HeadMeshSurfaceLayer):
    _type: ClassVar[str] = 'SampleMetadataInterpolatedSurface'
    _metadataKey: str
    """
    Which value in sample metadata to plot
    """
    _colorbarLabel: str | None = None
    _scalarAnnotations: dict[float, str] | None = None
    """
    In form accepted by pyvista.mapper.set_scalars(..., annotations=...).
    E.g.
        {50e-6: '50 uV', 1e-3: '1 mV'}
    """
    _relevantSampleDepth: str = 'target'
    """
    Which depth along sample entry vector to use for interpolation on surface.
    If interpolating onto cortical surface, probably want to use skin depth.
    If interpolating onto skin surface, probably want to use entry or coil depth.
    Or for any smooth-ish surface, can use "intersection" to autoset depth based on intersection with plotted surface.
    """
    _kernelSharpness: float = 0.5  # TODO: experiment to set more reasonable default
    _kernelRadius: float = 5.0

    _mesh: pv.PolyData | None = attrs.field(init=False, default=None)
    _scalarsKey: str = 'SampleMetadataInterpolated'
    """
    Where to save interpolated values within internal mesh object
    """
    # _scalarsOpacityKey: str | None = 'SampleMetadataInterpolatedOpacity'
    _scalarsOpacityKey: str | None = None

    _meshOpacityOutsideInterpolatedRegion: float = 1.

    _reinterpolationRateLimit: float = 10  # in Hz
    _needsReinterpolation: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)

    _cachedSampleIntersections: dict[str, np.ndarray] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._coordinator.session.samples.sigItemsChanged.connect(self._onSamplesChanged)

        asyncio.create_task(asyncTryAndLogExceptionOnError(self._loop_reinterpolate))

    async def _loop_reinterpolate(self):
        """
        Handle re-interpolation in an async loop to rate-limit (important when multiple samples are edited rapidly)
        """

        while True:
            logger.debug('Awaiting needsReinterpolation')
            await self._needsReinterpolation.wait()
            logger.debug('Sleeping before reinterpolating')
            await asyncio.sleep(1/self._reinterpolationRateLimit)
            if not self._needsReinterpolation.is_set():
                continue
            logger.debug('Reinterpolation loop')
            self._redraw(which='interpolateValues')

    def _redraw(self, which: tp.Union[tp.Optional[str], tp.List[str, ...]] = None):
        logger.debug(f'redraw {which}')
        if which == 'initSurf':
            assert self._mesh is None

            # override parent method to use sample metadata to interpolate surface
            self._mesh = pv.PolyData(getattr(self._coordinator.session.headModel, self._surfKey),
                                     deep=True)

            if True:  # TODO: troubleshoot to be able to set initValue to nan instead of zero
                # this seems to cause issues with colorbar scaling
                initValue = np.nan
            else:
                initValue = 0.
            initValues = np.full((self._mesh.n_points,), initValue)
            initValues[0] = -40  # TODO: debug, delete
            self._mesh[self._scalarsKey] = initValues

            actorKey = self._getActorKey('surf')

            if self._colorbarLabel is None:
                colorbarLabel = self._metadataKey
            else:
                colorbarLabel = self._colorbarLabel
            scalar_bar_args = dict(
                title=colorbarLabel,
                bold=True,
                title_font_size=12,
                label_font_size=10,
                vertical=True,
            )

            self._mesh.clear_data()  # TODO: determine if this is necessary

            self._interpolateValuesOntoMesh()

            if self._scalarsOpacityKey is not None:
                opacity = self._scalarsOpacityKey
            else:
                opacity = self._opacity

            self._actors[actorKey] = self._plotter.addMesh(mesh=self._mesh,
                                                            color=self._color,
                                                            nan_color=self._color,
                                                            nan_opacity=self._meshOpacityOutsideInterpolatedRegion,
                                                            scalars=self._scalarsKey,
                                                            scalar_bar_args=scalar_bar_args,
                                                            annotations=self._scalarAnnotations,
                                                            opacity=opacity,
                                                            specular=0.5,
                                                            diffuse=0.5,
                                                            ambient=0.5,
                                                            #smooth_shading=True,  # disabled since this breaks scalar value updates later
                                                            split_sharp_edges=True,
                                                            name=actorKey)

            #self._plotter.reset_scalar_bar_ranges([self._colorbarLabel])

            self._plotter.reset_camera_clipping_range()

            if False:
                # if not immmediately interpolating values above, need to do afterwards
                self._redraw(['updatePosition', 'interpolateValues'])
            else:
                self._redraw(['updatePosition'])

        elif which == 'queueInterpolateValues':
            self._needsReinterpolation.set()

        elif which == 'interpolateValues':

            if self._scalarsOpacityKey is None and False:

                self._interpolateValuesOntoMesh()

                # just call update rather than completely re-adding the mesh to plotter
                with self._plotter.allowNonblockingCalls():
                    # self._plotter.update_scalars(self._scalarsKey,
                    #                              mesh=self._mesh,
                    #                              render=True)
                    #self._plotter.reset_scalar_bar_ranges(scalarBarTitles=[self._colorbarLabel])
                    self._plotter.update()

            else:
                # current pyvista doesn't support dynamically changing color when a custom opacity is specified, so
                # need to fully re-add the mesh to the plotter

                actorKey = self._getActorKey('surf')

                assert actorKey in self._actors

                with self._plotter.allowNonblockingCalls():
                    self._plotter.remove_actor(self._actors[actorKey])
                    self._actors.pop(actorKey)

                if self._colorbarLabel is None:
                    colorbarLabel = self._metadataKey
                else:
                    colorbarLabel = self._colorbarLabel
                scalar_bar_args = dict(
                    title=colorbarLabel,
                    bold=True,
                    title_font_size=12,
                    label_font_size=10,
                    vertical=True,
                )

                self._interpolateValuesOntoMesh()

                if self._scalarsOpacityKey is not None:
                    opacity = self._scalarsOpacityKey
                else:
                    opacity = self._opacity

                self._actors[actorKey] = self._plotter.addMesh(mesh=self._mesh,
                                                               color=self._color,
                                                               nan_color=self._color,
                                                               scalars=self._scalarsKey,
                                                               scalar_bar_args=scalar_bar_args,
                                                               annotations=self._scalarAnnotations,
                                                               opacity=opacity,
                                                               specular=0.5,
                                                               diffuse=0.5,
                                                               ambient=0.5,
                                                               # smooth_shading=True,  # disabled since this breaks scalar value updates later
                                                               split_sharp_edges=True,
                                                               name=actorKey)

                with self._plotter.allowNonblockingCalls():
                    # self._plotter.reset_scalar_bar_ranges([self._colorbarLabel])

                    self._plotter.reset_camera_clipping_range()

        else:
            super()._redraw(which=which)

    def _interpolateValuesOntoMesh(self):
        self._needsReinterpolation.clear()

        allSamples = self._coordinator.session.samples.values()
        includeSamples = [sample for sample in allSamples if sample.isVisible]  # include only samples marked as visible

        coords = np.full((len(includeSamples), 3), np.nan)

        match self._relevantSampleDepth:
            case 'coil':
                for iSample, sample in enumerate(includeSamples):
                    if sample.coilToMRITransf is None:
                        continue
                    coords[iSample, :] = sample.coilToMRITransf[:3, 3].T

            case 'entry':
                for iSample, sample in enumerate(includeSamples):
                    if sample.coilToMRITransf is None:
                        continue
                    if sample.targetKey is None:
                        continue
                    target = self._coordinator.session.targets[sample.targetKey]
                    if target.entryCoordPlusDepthOffset is None or target.entryCoord is None:
                        continue
                    coilToEntryDepth = np.linalg.norm(target.entryCoordPlusDepthOffset - target.entryCoord)
                    coords[iSample, :] = applyTransform(sample.coilToMRITransf, np.asarray([0, 0, -coilToEntryDepth]), doCheck=False)

            case 'intersection':
                for iSample, sample in enumerate(includeSamples):
                    # ray tracing takes a non-negligible amount of time, and results don't change for most samples between updates, so cache results
                    try:
                        coords[iSample, :] = self._cachedSampleIntersections[sample.key]
                    except KeyError:
                        pass  # no cached value available
                    else:
                        continue

                    if sample.coilToMRITransf is None:
                        continue
                    rayPoints_coilSpace = np.asarray([[0, 0, 0,], [0, 0, -100]])
                    rayPoints_mriSpace = applyTransform(sample.coilToMRITransf, rayPoints_coilSpace, doCheck=False)

                    intersectionPoints, intersectionCells = self._mesh.ray_trace(
                        origin=rayPoints_mriSpace[0, :],
                        end_point=rayPoints_mriSpace[1, :],
                        first_point=True)
                    if len(intersectionPoints) == 0:
                        continue  # no intersection
                    coords[iSample, :] = intersectionPoints

                    self._cachedSampleIntersections[sample.key] = coords[iSample, :]

            case 'target':
                for iSample, sample in enumerate(includeSamples):
                    if sample.coilToMRITransf is None:
                        continue
                    if sample.targetKey is None:
                        continue
                    target = self._coordinator.session.targets[sample.targetKey]
                    if target.entryCoordPlusDepthOffset is None or target.targetCoord is None:
                        continue
                    coilToTargetDepth = np.linalg.norm(target.entryCoordPlusDepthOffset - target.targetCoord)
                    coords[iSample, :] = applyTransform(sample.coilToMRITransf, np.asarray([0, 0, -coilToTargetDepth]), doCheck=False)

            case _:
                raise NotImplementedError

        # TODO: cache this point cloud between calls and just update iteratively on changes

        values = np.full((coords.shape[0],), np.nan)
        for iSample, sample in enumerate(includeSamples):
            values[iSample] = sample.metadata.get(self._metadataKey, np.nan)

        if True:
            # remove samples with nan values
            indices = np.where(np.isnan(values))
            if len(indices) > 0:
                coords = np.delete(coords, indices, axis=0)
                values = np.delete(values, indices, axis=0)

        if len(coords) == 0:
            # nothing to interpolate
            newVals = np.full((self._mesh.n_points,), np.nan)
        else:

            pointCloud = pv.PointSet(coords)
            pointCloud[self._scalarsKey] = values

            self._mesh.clear_data()  # not sure why this is necessary, but otherwise the interpolation doesn't work
            logger.debug(f'Interpolating point cloud values {self._metadataKey} onto surface {self._surfKey}')
            tmpMesh = self._mesh.interpolate(pointCloud,
                                             null_value=np.nan,
                                             sharpness=self._kernelSharpness,
                                             radius=self._kernelRadius
                                             )
            logger.debug(f'Done interpolating')

            if self._scalarsKey not in tmpMesh.point_data:
                # this can happen if all interpolated values were nan
                newVals = np.full((self._mesh.n_points,), np.nan)
            else:
                newVals = tmpMesh[self._scalarsKey]

        if self._scalarsKey not in self._mesh.point_data or True:
            self._mesh[self._scalarsKey] = newVals
        else:
            self._mesh[self._scalarsKey][:] = newVals  # update existing array

        if self._scalarsOpacityKey is not None:
            # set opacity to 0 for nan values
            opacityVals = np.where(np.isnan(self._mesh[self._scalarsKey]), 0, self._opacity)
            if self._scalarsOpacityKey not in self._mesh.point_data:
                self._mesh[self._scalarsOpacityKey] = opacityVals
            else:
                self._mesh[self._scalarsOpacityKey][:] = opacityVals  # update existing array

        logger.debug(f'{np.isnan(self._mesh[self._scalarsKey]).sum()} / {self._mesh.n_points} points are nan')

    def _onSamplesChanged(self, changedKeys: tp.List[str], changedAttrs: tp.Optional[tp.List[str]]):
        logger.debug(f'_onSamplesChanged: {changedKeys} {changedAttrs}')
        if changedAttrs is not None:
            if 'isVisible' not in changedAttrs:
                if any(x in changedAttrs for x in ('coilToMRITransf', 'targetKey', 'metadata')):
                    anyIncludedSampleChanged = False
                    for sampleKey in changedKeys:
                        sample = self._coordinator.session.samples.get(sampleKey, None)
                        if sample is None:
                            continue
                        if sample.isVisible:
                            anyIncludedSampleChanged = True
                            break
                    if not anyIncludedSampleChanged:
                        # no attributes changed that would affect metadata interpolation
                        return
                else:
                    # no attributes changed that would affect metadata interpolation
                    return
        if len(self._cachedSampleIntersections) > 0:
            # clear relevant entries in cache
            for key in changedKeys:
                try:
                    self._cachedSampleIntersections.pop(key)
                except KeyError:
                    pass
        self._redraw(which='queueInterpolateValues')



"""
Adapted from https://stackoverflow.com/a/61274630/2388228
"""

from qtpy import QtWidgets, QtCore, QtGui


def preventAnnoyingScrollBehaviour(control: QtWidgets.QWidget) -> None:
    control.setFocusPolicy(QtCore.Qt.StrongFocus)
    control.installEventFilter(_MouseWheelWidgetAdjustmentGuard(control))


class _MouseWheelWidgetAdjustmentGuard(QtCore.QObject):
    def __init__(self, parent: QtWidgets.QWidget):
        super().__init__(parent)

    def eventFilter(self, o: QtCore.QObject, e: QtCore.QEvent) -> bool:
        widget: QtWidgets.QWidget = o
        if e.type() == QtCore.QEvent.Wheel and not widget.hasFocus():
            e.ignore()
            return True
        return super().eventFilter(o, e)



from __future__ import annotations

import asyncio
import typing as tp

import attrs
import logging
import numpy as np
import pandas as pd
import pylsl as lsl
from qtpy import QtWidgets

from .TriggerSourceSettingsWidget import TriggerSourceSettingsWidget
from NaviNIBS.Navigator.Model.Triggering import LSLTriggerSource, TriggerEvent, TriggerSource
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError
from NaviNIBS.util.lsl.LSLStreamSelector import LSLStreamSelector


logger = logging.getLogger(__name__)


@attrs.define
class LSLTriggerSourceSettingsWidget(TriggerSourceSettingsWidget[LSLTriggerSource]):
    _title: str = 'LSL trigger settings'
    _triggerSourceKey: str = 'LSLTriggerSource'
    _streamSelector: LSLStreamSelector = attrs.field(init=False)
    _inlet: tp.Optional[lsl.StreamInlet] = attrs.field(init=False, default=None)
    _inletConnectedEvent: asyncio.Event = attrs.field(init=False, factory=asyncio.Event)
    _pollPeriod: float = 0.05
    _pollTask: asyncio.Task = attrs.field(init=False)
    _lastTriggerTimePerAction: dict[str, tp.Optional[pd.Timestamp]] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

        self._wdgt.setLayout(QtWidgets.QFormLayout())

        if self.triggerSource is None:
            # create new empty LSL trigger source
            self.session.triggerSources.addItem(LSLTriggerSource(key=self._triggerSourceKey))
            assert self.triggerSource is not None

        initialSelectedStreamKey = self.triggerSource.streamKey

        self._streamSelector = LSLStreamSelector(selectedStreamKey=initialSelectedStreamKey)
        self._streamSelector.sigSelectedStreamKeyChanged.connect(self._onSelectedStreamKeyChanged)
        self._streamSelector.sigSelectedStreamAvailabilityChanged.connect(self._onSelectedStreamAvailabilityChanged)

        self.triggerSource.sigItemChanged.connect(self._onTriggerSourceItemChanged)

        self._wdgt.layout().addRow('Trigger stream', self._streamSelector.wdgt)

        self._pollTask = asyncio.create_task(asyncTryAndLogExceptionOnError(self._pollForData))

        if self.fallbackTriggerSource is not None:
            self.fallbackTriggerSource.isEnabled = True

    @property
    def fallbackTriggerSource(self) -> TriggerSource | None:
        if self.triggerSource.fallbackTriggerSourceKey is None:
            return None

        if self.session is not None and self.triggerSource.fallbackTriggerSourceKey in self.session.triggerSources:
            return self.session.triggerSources[self.triggerSource.fallbackTriggerSourceKey]

        return None

    def _disconnectInlet(self):
        logger.info(f'Disconnecting from LSL stream {self._streamSelector.selectedStreamKey}')
        self._inletConnectedEvent.clear()
        self._inlet.close_stream()
        self._inlet = None
        logger.debug(f'Disconnected from LSL stream {self._streamSelector.selectedStreamKey}')
        # TODO: maybe start and stop polling task when connecting / disconnecting inlet
        if self.fallbackTriggerSource is not None:
            self.fallbackTriggerSource.isEnabled = True

    def _connectInlet(self):
        assert self._inlet is None
        if not self._streamSelector.selectedStreamIsAvailable:
            logger.info('Selected trigger stream is not available. Skipping attempt to connect')
            return

        isFixedRate = self._streamSelector.selectedStreamInfo.nominal_srate() > 0
        if isFixedRate:
            logger.warning('Fixed rate streams not currently supported. Not connecting inlet.')
            return

        logger.info(f'Connecting to LSL stream {self._streamSelector.selectedStreamKey}')
        self._inlet = lsl.StreamInlet(self._streamSelector.selectedStreamInfo,
                                      max_buflen=30,
                                      processing_flags=lsl.proc_ALL,
                                      recover=False)
        self._inletConnectedEvent.set()
        logger.debug(f'Connected to LSL stream {self._streamSelector.selectedStreamKey}')

        if self.fallbackTriggerSource is not None:
            self.fallbackTriggerSource.isEnabled = False

    async def _pollForData(self):
        while True:
            await self._inletConnectedEvent.wait()

            try:
                chunk, timestamps = self._inlet.pull_chunk(
                    timeout=0.,
                    max_samples=10
                )
            except lsl.LostError as e:
                logger.info(f'Previously connected stream inlet {self._streamSelector.selectedStreamKey} is no longer available')
                self._streamSelector.markStreamAsLost(streamKey=self._streamSelector.selectedStreamKey)
                self._disconnectInlet()
                continue

            chunk = np.asarray(chunk)
            if chunk.ndim == 2 and chunk.shape[1] == 1:
                chunk.resize(chunk.shape[0])

            if len(chunk) == 0:
                # no samples
                await asyncio.sleep(self._pollPeriod)
                continue

            evtTimes, evtData, evtIndices = self._getRelevantEvents(timestamps, chunk)

            if len(evtTimes) == 0:
                # no relevant events
                continue

            for evtTime, evtDat in zip(evtTimes, evtData):
                action = None
                if self.triggerSource.triggerEvents is not None:
                    action = self.triggerSource.triggerEvents.get(evtDat, None)
                if action is None:
                    action = self.triggerSource.defaultAction

                try:
                    metadata = dict(originalType=int(evtDat))
                except ValueError as e:
                    metadata = dict()

                if self.triggerSource.triggerValueIsEpochID:
                    metadata['epochID'] = int(evtDat)
                    # TODO: assert that epochID is actually unique (to catch issues where the selected stream
                    # is actually sending non-unique events rather than unique epochIDs)

                triggerEvt = TriggerEvent(
                    type=action,
                    time=pd.Timestamp.now() + pd.Timedelta(seconds=(evtTime - lsl.local_clock())),  # convert from lsl time to pandas timestamp
                    metadata=metadata
                )
                if self._lastTriggerTimePerAction.get(action, None) is not None and (triggerEvt.time - self._lastTriggerTimePerAction[action]).total_seconds() < self.triggerSource.minInterTriggerPeriod:
                    logger.debug('Ignoring trigger that occured too quickly after previous')
                    continue
                self.triggerSource.trigger(triggerEvt)
                self._lastTriggerTimePerAction[action] = triggerEvt.time

    def _getRelevantEvents(self, evtTimes: list[float], evtData: np.ndarray) -> tuple[list[float], list[np.ndarray], np.ndarray]:
        if self.triggerSource.triggerEvents is not None and len(self.triggerSource.triggerEvents) > 0:
            # only trigger for specific events of interest
            relevantEvtIndices = np.isin(evtData, self.triggerSource.triggerEvents)
        else:
            # trigger for any event
            if False:
                relevantEvtIndices = np.full(evtData.shape, True)
            else:
                # (except for off events like 0 and False)
                relevantEvtIndices = np.logical_not(np.isin(evtData, ('0', 'False')))

        if not np.any(relevantEvtIndices):
            return [], [], []

        relevantEvtTimes = []
        relevantEvtData = []
        for evtIndex in np.nonzero(relevantEvtIndices)[0]:
            relevantEvtTimes.append(evtTimes[evtIndex])
            relevantEvtData.append(evtData[evtIndex])
        return relevantEvtTimes, relevantEvtData, relevantEvtIndices

    def _connectOrDisconnectAsNeeded(self):
        if self._inlet is None and (self.triggerSource.isEnabled and self._streamSelector.selectedStreamIsAvailable):
            self._connectInlet()
        elif self._inlet is not None and (not self.triggerSource.isEnabled or not self._streamSelector.selectedStreamIsAvailable):
            self._disconnectInlet()

    def _onSelectedStreamKeyChanged(self, newKey: str):
        if self.session is None:
            # not yet initialized
            return
        if self._inlet is not None:
            self._disconnectInlet()
        self.triggerSource.streamKey = newKey
        self._connectOrDisconnectAsNeeded()

    def _onSelectedStreamAvailabilityChanged(self):
        self._connectOrDisconnectAsNeeded()

    def _onTriggerSourceItemChanged(self, key: str, whichAttrs: list[str] | None = None):
        if whichAttrs is None or any(x in whichAttrs for x in ('fallbackTriggerSourceKey', 'isEnabled')):
            if self.fallbackTriggerSource is not None:
                self.fallbackTriggerSource.isEnabled = self._inlet is None and self.triggerSource.isEnabled

        if whichAttrs is None or 'streamKey' in whichAttrs:
            self._streamSelector.selectedStreamKey = self.triggerSource.streamKey

        if whichAttrs is None or 'isEnabled' in whichAttrs:
            self._connectOrDisconnectAsNeeded()

    def _onSessionTriggerSettingChanged(self, triggerSourceKeys: list[str], attribs: tp.Optional[list[str]]):
        if self._triggerSourceKey not in triggerSourceKeys:
            return  # ignore other trigger sources
        self._streamSelector.selectedStreamKey = self.triggerSource.streamKey
        # TODO: handle any other change updates as needed

    def _onSessionSet(self):
        if self.triggerSource is None:
            # create new trigger source settings
            self.session.triggerSources[self._triggerSourceKey] = LSLTriggerSource()
        self._streamSelector.selectedStreamKey = self.triggerSource.streamKey
        self._session.triggerSources.sigItemsChanged.connect(self._onSessionTriggerSettingChanged)
